#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# EazeBot
# Free python/telegram bot for easy execution and surveillance of crypto trading plans on multiple exchanges.
# Copyright (C) 2018
# Marcel Beining <marcel.beining@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This module contains all functions necessary for starting the Bot"""


#%% import modules
import logging
import logging.handlers  # necessary if run as main script and not interactively...dunno why
import re
import time
import datetime as dt
import json
import dill
import requests
import base64
from copy import deepcopy
from shutil import copy2
from collections import defaultdict
import os
from telegram import (ReplyKeyboardMarkup,InlineKeyboardMarkup,InlineKeyboardButton,bot)
from telegram.ext import (Updater, CommandHandler, MessageHandler, Filters, RegexHandler,
                          ConversationHandler,CallbackQueryHandler)
from telegram.error import BadRequest

if __name__ == '__main__' or os.path.isfile('tradeHandler.py'):
    from tradeHandler import tradeHandler
else:
    from eazebot.tradeHandler import tradeHandler

logFileName = 'telegramEazeBot'
MAINMENU,SETTINGS,SYMBOL,NUMBER,TIMING,INFO = range(6)

logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
rootLogger = logging.getLogger()
rootLogger.handlers = []  # delete old handlers in case bot is restarted but not python kernel
rootLogger.setLevel('INFO') #DEBUG
fileHandler = logging.handlers.RotatingFileHandler("{0}/{1}.log".format(os.getcwd(), logFileName),maxBytes=1000000, backupCount=5)
fileHandler.setFormatter(logFormatter)
rootLogger.addHandler(fileHandler)
consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logFormatter)
rootLogger.addHandler(consoleHandler)


with open(os.path.join(os.path.dirname(__file__),'version.txt')) as fh:
    thisVersion = re.search('(?<=version = )[0-9\.]+',str(fh.read())).group(0)

#%% init menues
mainMenu = [['Status of Trade Sets', 'New Trade Set','Trade History'],['Check Balance','Bot Info'],['Add/update exchanges (APIs.json)','Settings']]
markupMainMenu = ReplyKeyboardMarkup(mainMenu)#, one_time_keyboard=True)

tradeSetMenu = [['Add buy position', 'Add sell position','Add initial coins'],
                  ['Add stop-loss','Show trade set','Done','Cancel']]
markupTradeSetMenu = ReplyKeyboardMarkup(tradeSetMenu,one_time_keyboard=True)


#%% init base variables
__config__ = {}
job_queue = []

## define  helper functions
def copyJSON(folderName=os.getcwd(),force=0):
    if force == 0 and os.path.isfile(os.path.join(folderName,'botConfig.json')):
        logging.warning('botConfig.json already exists in\n%s\nUse copyJSON(targetfolder,force=1) or copyJSON(force=1) to overwrite both (!) JSONs'%folderName)
    else:
        copy2(os.path.join(os.path.dirname(__file__),'botConfig.json'),folderName)
    if force == 0 and os.path.isfile(os.path.join(folderName,'APIs.json')):
        logging.warning('APIs.json already exists in\n%s\nUse copyJSON(targetfolder,force=1) or copyJSON(force=1) to overwrite both (!) JSONs'%folderName)
    else:  
        copy2(os.path.join(os.path.dirname(__file__),'APIs.json'),folderName)
    copy2(os.path.join(os.path.dirname(__file__),'startBotScript.py'),folderName)
    copy2(os.path.join(os.path.dirname(__file__),'startBot.bat'),folderName)
    copy2(os.path.join(os.path.dirname(__file__),'updateBot.bat'),folderName)
    logging.info('botConfig.json and APIs.json successfully copied to\n%s\nPlease open and configure these files before running the bot'%folderName)
        
def broadcastMsg(bot,userId,msg,level='info'):
    # put msg into log with userId
    getattr(rootLogger,level.lower())('User %d: %s'%(userId,msg))
    if __config__['debug'] or level.lower() != 'debug':
        # return msg to user
        count = 0
        while count < 5:
            try:
                bot.send_message(chat_id=userId, text=level + ': ' + msg, parse_mode='markdown')
                break
            except TypeError as e:
                pass            
            except:
                count += 1
                logging.warning('Some connection (?) error occured when trying to send a telegram message. Retrying..')
                time.sleep(1)
                continue
        if count >= 5:
            logging.error('Could not send message to bot')

def noncomprende(bot, update,what):
    if what == 'unknownCmd':
        txt = "Sorry, I didn't understand that command."
    elif what == 'wrongSymbolFormat':
        txt = "Sorry, the currency pair is not in the form COINA/COINB. Retry!"
    elif what == 'noNumber':
        txt = "Sorry, you did not enter a number! Retry!"
    elif what == 'noValueRequested':
        txt = "Sorry, I did not ask for anything at the moment, and unfortunately I have no KI (yet) ;-)"
    else:
        txt = what
    while True:
        try:
            bot.send_message(chat_id=update.message.chat_id, text=txt)
            break
        except:
            continue

def getCName(symbol,which=0):
    if which == 0:
        return re.search('^\w+(?=/)',symbol).group(0)
    else:
        return re.search('(?<=/)\w+$',symbol).group(0)
             
def receivedInfo(bot,update,user_data):    
    if len(user_data['lastFct']) > 0:
        return user_data['lastFct'].pop()(update.message.text)
    else:
        bot.send_message(user_data['chatId'],'Unknown previous error, returning to main menu')
        return MAINMENU

def receivedFloat(bot,update,user_data): 
    if len(user_data['lastFct']) > 0:
        return user_data['lastFct'].pop()(float(update.message.text))
    else:
        bot.send_message(user_data['chatId'],'Unknown previous error, returning to main menu')
        return MAINMENU
      
        
## define menu function
def startCmd(bot, update,user_data):
    # initiate user_data if it does not exist yet
    if update.message.from_user.id not in __config__['telegramUserId']:
        bot.send_message(update.message.from_user.id,'Sorry your Telegram ID (%d) is not recognized! Bye!'%update.message.from_user.id)
        logging.warning('Unknown user %s %s (username: %s, id: %s) tried to start the bot!'%(update.message.from_user.first_name,update.message.from_user.last_name,update.message.from_user.username,update.message.from_user.id))
        return
    else:
        logging.info('User %s %s (username: %s, id: %s) (re)started the bot'%(update.message.from_user.first_name,update.message.from_user.last_name,update.message.from_user.username,update.message.from_user.id))
    if user_data:
        washere = 'back '
        user_data.update({'lastFct':[],'whichCurrency':0,'tempTradeSet':[None,None,None],'messages':{'status':[],'dialog':[],'botInfo':[],'settings':[],'history':[]}})
    else:
        washere = ''
        user_data.update({'chatId':update.message.chat_id,'exchanges':{},'trade':{},'settings':{'fiat':[],'showProfitIn':None},'lastFct':[],'whichCurrency':0,'tempTradeSet':[None,None,None],'messages':{'status':[],'dialog':[],'botInfo':[],'settings':[]}})
    deleteMessages(user_data)
    bot.send_message(user_data['chatId'],
        "Welcome %s%s to the EazeBot! You are in the main menu."%(washere,update.message.from_user.first_name),
        reply_markup=markupMainMenu)
    return MAINMENU

def makeTSInlineKeyboard(exch,iTS):
    button_list = [[
        InlineKeyboardButton("Edit Set", callback_data='2|%s|%s'%(exch,iTS)),
        InlineKeyboardButton("Delete/SellAll", callback_data='3|%s|%s'%(exch,iTS))]]
    return InlineKeyboardMarkup(button_list)

def buttonsEditTS(ct,uidTS,mode='full'):
    exch = ct.exchange.name.lower()
    buttons = [[InlineKeyboardButton("Add buy level",callback_data='2|%s|%s|buyAdd|chosen'%(exch,uidTS)),InlineKeyboardButton("Add sell level",callback_data='2|%s|%s|sellAdd|chosen'%(exch,uidTS))]]
    for i,trade in enumerate(ct.tradeSets[uidTS]['InTrades']):
        if trade['oid'] == 'filled':
            buttons.append([InlineKeyboardButton("Readd BuyOrder from level #%d"%i,callback_data='2|%s|%s|buyReAdd%d|chosen'%(exch,uidTS,i))])
        buttons.append([InlineKeyboardButton("Delete Buy level #%d"%i,callback_data='2|%s|%s|BLD%d|chosen'%(exch,uidTS,i))])
    for i,trade in enumerate(ct.tradeSets[uidTS]['OutTrades']):
        if trade['oid'] == 'filled':
            buttons.append([InlineKeyboardButton("Readd SellOrder from level #%d"%i,callback_data='2|%s|%s|sellReAdd%d|chosen'%(exch,uidTS,i))])
        buttons.append([InlineKeyboardButton("Delete Sell level #%d"%i,callback_data='2|%s|%s|SLD%d|chosen'%(exch,uidTS,i))])
    if mode == 'full':
        buttons.append([InlineKeyboardButton("Set/Change SL",callback_data='2|%s|%s|SLM'%(exch,uidTS))])
    elif mode == 'init':
        buttons.append([InlineKeyboardButton("Add initial coins",callback_data='2|%s|%s|AIC|chosen'%(exch,uidTS)),InlineKeyboardButton("Add/change SL",callback_data='2|%s|%s|SLC|chosen'%(exch,uidTS))])
    buttons.append([InlineKeyboardButton("%s trade set"%('Deactivate' if ct.tradeSets[uidTS]['active'] else 'Activate'),callback_data='2|%s|%s|%s|chosen'%(exch,uidTS,'TSstop' if ct.tradeSets[uidTS]['active'] else 'TSgo')),InlineKeyboardButton("Delete trade set",callback_data='3|%s|%s|ok|no|chosen'%(exch,uidTS))])
    if mode == 'full':
        buttons.append([InlineKeyboardButton("Back",callback_data='2|%s|%s|back|chosen'%(exch,uidTS))])
    return buttons

def buttonsSL(ct,uidTS):
    exch = ct.exchange.name.lower()   
    buttons = [[InlineKeyboardButton("Set SL Break Even",callback_data='2|%s|%s|SLBE|chosen'%(exch,uidTS))],[InlineKeyboardButton("Change/Delete SL",callback_data='2|%s|%s|SLC|chosen'%(exch,uidTS))],[InlineKeyboardButton("Set daily-close SL",callback_data='2|%s|%s|DCSL|chosen'%(exch,uidTS))],[InlineKeyboardButton("Set weekly-close SL",callback_data='2|%s|%s|WCSL|chosen'%(exch,uidTS))]]
    if ct.numBuyLevels(uidTS,'notfilled') == 0:  # only show trailing SL option if all buy orders are filled
        buttons.append([InlineKeyboardButton("Set trailing SL",callback_data='2|%s|%s|TSL|chosen'%(exch,uidTS))])
    buttons.append([InlineKeyboardButton("Back",callback_data='2|%s|%s|back|chosen'%(exch,uidTS))])
    return buttons

def buttonsEditTSH(ct):
    exch = ct.exchange.name.lower()   
    ct.resetTradeHistory
    return InlineKeyboardMarkup([[InlineKeyboardButton("Clear Trade History",callback_data='resetTSH|%s|XXX'%(exch))]])


def deleteMessages(user_data,typ='all',onlyForget=False):
    if isinstance(typ,str):
        if typ == 'all':
            typ = list(user_data['messages'].keys())
        elif not isinstance(typ,list):
            typ = [typ]
    for t in typ:
        if not onlyForget:
            for msg in user_data['messages'][t]:
                try:
                    msg.delete()
                except:
                    pass
        user_data['messages'][t] = []
    return 1
    
def printTradeStatus(bot,update,user_data,onlyThisTs=None):
    count = 0
    deleteMessages(user_data,'status')
    for iex,ex in enumerate(user_data['trade']):
        ct = user_data['trade'][ex]
        if onlyThisTs is not None and onlyThisTs not in ct.tradeSets:
            continue
        count = 0       
        for iTs in ct.tradeSets:
            try:  # catch errors in order to be able to see the statuses of other exchanges, if one exchange has a problem
                ts = ct.tradeSets[iTs]
                if onlyThisTs is not None and onlyThisTs != iTs:
                    continue
                if ts['virgin']:
                    markup = InlineKeyboardMarkup(buttonsEditTS(ct,iTs,mode='init'))
                else:
                    markup = makeTSInlineKeyboard(ex,iTs)
                count += 1
                user_data['messages']['status'].append(bot.send_message(user_data['chatId'],ct.getTradeSetInfo(iTs,user_data['settings']['showProfitIn']),reply_markup=markup,parse_mode='markdown'))
            except Exception as e:
                logging.error(str(e))
                pass
        if count == 0:
            user_data['messages']['status'].append(bot.send_message(user_data['chatId'],'No Trade sets found on %s'%ex))
    if len(user_data['trade']) == 0:
        user_data['messages']['status'].append(bot.send_message(user_data['chatId'],'No exchange found to check trade sets'))
    return MAINMENU 

def printTradeHistory(bot,update,user_data):
    deleteMessages(user_data,'history') #%!!!
    for iex,ex in enumerate(user_data['trade']):
        ct = user_data['trade'][ex]
        user_data['messages']['history'].append(bot.send_message(user_data['chatId'],ct.getTradeHistory(),parse_mode='markdown',reply_markup = buttonsEditTSH(ct) ))
    return MAINMENU 
    
def checkBalance(bot,update,user_data,exchange=None):
    if exchange:
        ct = user_data['trade'][exchange]
        ct.updateBalance()
        if ct.exchange.has['fetchTickers']:
            tickers = ct.safeRun(ct.exchange.fetchTickers)
            func = lambda sym: tickers[sym] if sym in tickers else ct.safeRun(lambda : ct.exchange.fetchTicker(sym))  # includes a hot fix for some ccxt problems
        else:
            func = lambda sym: ct.safeRun(lambda : ct.exchange.fetchTicker(sym))
        coins = list(ct.balance['total'].keys())
        string = '*Balance on %s (>%g BTC):*\n'%(exchange,__config__['minBalanceInBTC'])
        for c in coins:
            BTCpair = '%s/BTC'%c
            BTCpair2 = 'BTC/%s'%c
            if ct.balance['total'][c] > 0:
                if c == 'BTC' and ct.balance['total'][c] > __config__['minBalanceInBTC']:
                    string += '*%s:* %s _(free: %s)_\n'%(c, ct.cost2Prec('ETH/BTC',ct.balance['total'][c]), ct.cost2Prec('ETH/BTC',ct.balance['free'][c]))
                elif BTCpair2 in ct.exchange.symbols and ct.balance['total'][c]/func(BTCpair2)['last'] > __config__['minBalanceInBTC'] :
                    string += '*%s:* %s _(free: %s)_\n'%(c, ct.cost2Prec(BTCpair2,ct.balance['total'][c]), ct.cost2Prec(BTCpair2,ct.balance['free'][c]))
                elif BTCpair in ct.exchange.symbols and func(BTCpair)['last']*ct.balance['total'][c] > __config__['minBalanceInBTC']:
                    string += '*%s:* %s _(free: %s)_\n'%(c, ct.amount2Prec(BTCpair,ct.balance['total'][c]), ct.amount2Prec(BTCpair,ct.balance['free'][c]))
                
        bot.send_message(user_data['chatId'],string,parse_mode='markdown')
    else:
        deleteMessages(user_data,'dialog')
        user_data['lastFct'].append(lambda res: checkBalance(bot,update,user_data,res))
        # list all available exanches for choosing
        exchs = [ct.exchange.name for _,ct in user_data['trade'].items()]
        buttons = [[InlineKeyboardButton(exch,callback_data='chooseExch|%s|xxx'%(exch.lower()))] for exch in sorted(exchs)] + [[InlineKeyboardButton('Cancel',callback_data='chooseExch|xxx|xxx|cancel')]]
        user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'For which exchange do you want to see your balance?',reply_markup=InlineKeyboardMarkup(buttons)))
            
    
def createTradeSet(bot,update,user_data,exchange=None,symbol=None):
    # check if user is registered and has any authenticated exchange
    if 'trade' in user_data and len(user_data['trade'])>0:
        # check if exchange was already chosen
        if exchange:
            ct = user_data['trade'][exchange]
            if symbol and symbol.upper() in ct.exchange.symbols:
                deleteMessages(user_data,'dialog')
                symbol = symbol.upper()                    
                ts, uidTS = ct.initTradeSet(symbol)
                ct.updateBalance()
                user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'Thank you, now let us begin setting the trade set'))
                printTradeStatus(bot,update,user_data,uidTS)
                return MAINMENU
            else:
                if symbol:
                    text = 'Symbol %s was not found on exchange %s'%(symbol,exchange)
                else:
                    text = 'Please specify your trade set now. First: Which currency pair do you want to trade? (e.g. ETH/BTC)'
                user_data['lastFct'].append(lambda res: createTradeSet(bot,update,user_data,exchange,res))
                user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],text,reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton('List all pairs on %s'%exchange,callback_data='showSymbols|%s|%s'%(exchange,'xxx')),InlineKeyboardButton('Cancel',callback_data='blabla|cancel')]])))
                return SYMBOL
        else:
            user_data['lastFct'].append(lambda res: createTradeSet(bot,update,user_data,res))
            # list all available exanches for choosing
            exchs = [ct.exchange.name for _,ct in user_data['trade'].items()]
            buttons = [[InlineKeyboardButton(exch,callback_data='chooseExch|%s|xxx'%(exch.lower()))] for exch in sorted(exchs)] + [[InlineKeyboardButton('Cancel',callback_data='chooseExch|xxx|xxx|cancel')]]
            user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'For which of your authenticated exchanges do you want to add a trade set?',reply_markup=InlineKeyboardMarkup(buttons)))
    else:
        user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'No authenticated exchanges found for your account! Please click "Add exchanges"'))
        return MAINMENU


def askAmount(user_data,exch,uidTS,direction,botOrQuery):
    ct = user_data['trade'][exch]
    ts = ct.tradeSets[uidTS]
    coin = ts['coinCurrency']
    currency = ts['baseCurrency']
    if direction=='sell':
        # free balance is coins available in trade set minus coins that will be sold plus coins that will be bought
        bal = ts['coinsAvail'] - ct.sumSellAmounts(uidTS,'notinitiated') + ct.sumBuyAmounts(uidTS,'notfilled',subtractFee = True)
        if user_data['whichCurrency']==0:
            bal = ct.amount2Prec(ts['symbol'],bal)
            cname = coin
            action = 'sell'
            balText = 'available %s [fee subtracted] is'%coin
        else:
            bal = ct.cost2Prec(ts['symbol'],bal*user_data['tempTradeSet'][0])
            cname = currency
            action = 'receive'
            balText = 'return from available %s [fee subtracted] would be'%coin
    elif direction == 'buy':
        # free balance is free currency minus cost for coins that will be bought
        bal = ct.getFreeBalance(currency) - ct.sumBuyCosts(uidTS,'notinitiated')
        if user_data['whichCurrency']==0:
            bal = ct.amount2Prec(ts['symbol'],bal/user_data['tempTradeSet'][0])
            cname = coin
            action = 'buy'
            balText = 'possible buy amount from your remaining free balance is'
        else:
            bal = ct.cost2Prec(ts['symbol'],bal)
            cname = currency
            action = 'use'
            balText = 'remaining free balance is'
    else:
        raise ValueError('Unknown direction specification')
    
    text = "What amount of %s do you want to %s (%s ~%s)?"%(cname,action,balText,bal)
    if isinstance(botOrQuery,bot.Bot):
        user_data['messages']['dialog'].append(botOrQuery.send_message(user_data['chatId'],text,reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton("Toggle currency", callback_data='toggleCurrency|%s|%s|%s'%(exch,uidTS,direction))],[InlineKeyboardButton("Cancel", callback_data='askAmount|cancel')]])))
    else:
        botOrQuery.edit_message_text(text,reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton("Toggle currency", callback_data='toggleCurrency|%s|%s|%s'%(exch,uidTS,direction))],[InlineKeyboardButton("Cancel", callback_data='askAmount|cancel')]]))
        botOrQuery.answer('Currency switched')
    
def addInitBalance(bot,user_data,exch,uidTS,inputType=None,response=None,fct = None):
    ct = user_data['trade'][exch]
    if inputType is None:
        user_data['lastFct'].append(lambda res : addInitBalance(bot,user_data,exch,uidTS,'initCoins',res,fct))
        user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],"You already have %s that you want to add to the trade set? How much is it (found %.5g free %s on %s)?"%(ct.tradeSets[uidTS]['coinCurrency'],ct.getFreeBalance(ct.tradeSets[uidTS]['coinCurrency']),ct.tradeSets[uidTS]['coinCurrency'],exch), reply_markup = InlineKeyboardMarkup([[InlineKeyboardButton("Cancel", callback_data='addInitBal|cancel')]])))
        return NUMBER
    elif inputType == 'initCoins':
        user_data['tempTradeSet'][0] = response
        user_data['lastFct'].append(lambda res : addInitBalance(bot,user_data,exch,uidTS,'initPrice',res,fct))
        user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],"What was the average price (%s) you bought it for? Type 0 if received for free and a negative number if you do not know?"%ct.tradeSets[uidTS]['symbol'], reply_markup = InlineKeyboardMarkup([[InlineKeyboardButton("Cancel", callback_data='addInitBal|cancel')]])))
        return NUMBER
    elif inputType == 'initPrice':
        deleteMessages(user_data,'dialog')
        if response >= 0:
            user_data['tempTradeSet'][1] = response
        addPos(bot,user_data,exch,uidTS,'init',fct)
        return MAINMENU

def addPos(bot,user_data,exch,uidTS,direction,fct=None):
    ct = user_data['trade'][exch]
    try:
        if direction == 'buy':
            ct.addBuyLevel(uidTS,user_data['tempTradeSet'][0],user_data['tempTradeSet'][1],user_data['tempTradeSet'][2])
        elif direction == 'sell':
            ct.addSellLevel(uidTS,user_data['tempTradeSet'][0],user_data['tempTradeSet'][1])
        else:
            ct.addInitCoins(uidTS,user_data['tempTradeSet'][0],user_data['tempTradeSet'][1])
    except Exception as e:
        broadcastMsg(bot,user_data['chatId'],str(e),'error')
    user_data['tempTradeSet'] = [None,None,None]
    if fct:
        fct()
        
def askPos(bot,user_data,exch,uidTS,direction,applyFct=None,inputType=None,response=None):
    ct = user_data['trade'][exch]
    symbol = ct.tradeSets[uidTS]['symbol']
    if inputType is None:
        user_data['tempTradeSet'] = [None,None,None]
        user_data['lastFct'].append(lambda res : askPos(bot,user_data,exch,uidTS,direction,applyFct,'price',res))
        user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],"At which price do you want to %s %s"%(direction,symbol), reply_markup = InlineKeyboardMarkup([[InlineKeyboardButton("Cancel", callback_data='askPos|cancel')]])))
        return NUMBER
    elif inputType == 'price':        
        if response == 0:
            user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],"Zero not allowed"))
            return NUMBER
        response = float(user_data['trade'][exch].exchange.priceToPrecision(symbol,response))
        user_data['tempTradeSet'][0] = response
        user_data['lastFct'].append(lambda res : askPos(bot,user_data,exch,uidTS,direction,applyFct,'amount',res))
        askAmount(user_data,exch,uidTS,direction,bot)
        return NUMBER
    elif inputType == 'amount':
        if user_data['whichCurrency']==1:
            response =response/user_data['tempTradeSet'][0]
        response = float(user_data['trade'][exch].exchange.amountToPrecision(symbol,response))
        user_data['tempTradeSet'][1] = response
        if direction == 'buy':
            user_data['lastFct'].append(lambda res : askPos(bot,user_data,exch,uidTS,direction,applyFct,'candleAbove',res))
            user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'Do you want to make this a timed buy (buy only if daily candle closes above X)',reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton("Yes", callback_data='Yes'),InlineKeyboardButton("No", callback_data='No')], [InlineKeyboardButton("Cancel", callback_data='askPos|cancel')]])))
            return TIMING    
        else:
            inputType = 'apply'
    if inputType == 'candleAbove':
        user_data['tempTradeSet'][2] = response
        inputType = 'apply'
    if inputType == 'apply':
        deleteMessages(user_data,'dialog')
        if applyFct == None:
            return addPos(bot,user_data,exch,direction)
        else:
            return applyFct()

def addExchanges(bot,update,user_data):
    idx = [i for i, x in enumerate(__config__['telegramUserId']) if x==user_data['chatId']][0]+1
    if idx == 1:
        with open("APIs.json", "r") as fin:
            APIs = json.load(fin)   
    else:
        with open("APIs%d.json"%idx, "r") as fin:
            APIs = json.load(fin)   
    keys = list(APIs.keys())
    hasKey = [re.search('(?<=^apiKey).*',val).group(0) for val in keys if re.search('(?<=^apiKey).*',val,re.IGNORECASE) is not None ]
    hasSecret = [re.search('(?<=^apiSecret).*',val).group(0) for val in keys if re.search('(?<=^apiSecret).*',val,re.IGNORECASE) is not None ]
    hasUid = [re.search('(?<=^apiUid).*',val).group(0) for val in keys if re.search('(?<=^apiUid).*',val,re.IGNORECASE) is not None ]
    hasPassword = [re.search('(?<=^apiPassword).*',val).group(0) for val in keys if re.search('(?<=^apiPassword).*',val,re.IGNORECASE) is not None ]
    availableExchanges = set(hasKey).intersection(set(hasSecret))
    if len(availableExchanges) > 0:
        logging.info('Found exchanges %s with keys %s, secrets %s, uids %s, password %s'%(availableExchanges,hasKey,hasSecret,hasUid,hasPassword))
        authenticatedExchanges = []
        for a in availableExchanges:
            exch = a.lower()
            exchParams = {'key': APIs['apiKey%s'%a] , 'secret': APIs['apiSecret%s'%a]}
            if a in hasUid:
                exchParams['uid'] = APIs['apiUid%s'%a]
            if a in hasPassword:
                exchParams['password'] = APIs['apiPassword%s'%a]
            # if no tradeHandler object has been created yet, create one, but also check for correct authentication
            if exch not in user_data['trade']:
                userId = user_data['chatId']
                user_data['trade'][exch] = tradeHandler(exch,**exchParams,messagerFct = lambda a,b='info': broadcastMsg(bot,userId,a,b),logger=rootLogger)
            else:
                user_data['trade'][exch].updateKeys(**exchParams)
            if not user_data['trade'][exch].authenticated and len(user_data['trade'][exch].tradeSets) == 0:
                user_data['trade'].pop(exch)
            else:
                authenticatedExchanges.append(a)
        bot.send_message(user_data['chatId'],'Exchanges %s added/updated'%authenticatedExchanges)
    else:
        bot.send_message(user_data['chatId'],'No exchange found to add')    
#        if update is not None:
#            return MAINMENU
    oldExchanges = set(ct.exchange.name for _,ct in user_data['trade'].items()) - availableExchanges
    removedExchanges = []
    for exch in oldExchanges:
        if len(user_data['trade'][exch].tradeSets) == 0:
            user_data['trade'].pop(exch)
            removedExchanges.append(exch)
    if len(removedExchanges) > 0:
        bot.send_message(user_data['chatId'],'Old exchanges %s with no tradeSets removed'%authenticatedExchanges)

def getRemoteVersion():
    remoteTxt = base64.b64decode(requests.get('https://api.github.com/repos/MarcelBeining/eazebot/contents/eazebot/version.txt').json()['content'])
    remoteVersion = re.search('(?<=version = )[0-9\.]+',str(remoteTxt)).group(0)
    remoteVersionCommit = [val['commit']['url'] for val in requests.get('https://api.github.com/repos/MarcelBeining/EazeBot/tags').json() if val['name']=='EazeBot_%s'%remoteVersion ][0]
    return (remoteVersion,requests.get(remoteVersionCommit).json()['commit']['message'])
        
def botInfo(bot,update,user_data):
    deleteMessages(user_data,'botInfo')
    string = '<b>******** EazeBot (v%s) ********</b>\n<i>Free python/telegram bot for easy execution and surveillance of crypto trading plans on multiple exchanges</i>\n'%thisVersion
    remoteVersion, versionMessage = getRemoteVersion()
    if remoteVersion != thisVersion and all([int(a) >= int(b) for a,b in zip(remoteVersion.split('.'),thisVersion.split('.'))]):
        string += '\n<b>There is a new version of EazeBot available on git (v%s) with these changes:\n%s\n</b>\n'%(remoteVersion,versionMessage)
    string+='\nReward my efforts on this bot by donating some cryptos!'
    user_data['messages']['botInfo'].append(bot.send_message(user_data['chatId'],string,parse_mode='html',reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton('Donate',callback_data='1|xxx|xxx')]])))
    return MAINMENU
    
def doneCmd(bot,update,user_data):
    updater.stop()
    bot.send_message(user_data['chatId'],"Bot stopped! Trades are not updated until starting again! See you soon %s! Start me again using /start anytime you want"%(update.message.from_user.first_name))
    logging.info('User %s (id: %s) ended the bot'%(update.message.from_user.first_name,update.message.from_user.id))
    
    
# job functions   
def checkForUpdates(bot,job):
    updater = job.context
    remoteVersion, versionMessage = getRemoteVersion()
    if remoteVersion != thisVersion and all([int(a) >= int(b) for a,b in zip(remoteVersion.split('.'),thisVersion.split('.'))]):
        for user in updater.dispatcher.user_data:
            if 'chatId' in updater.dispatcher.user_data[user]:
                updater.dispatcher.user_data[user]['messages']['botInfo'].append(bot.send_message(updater.dispatcher.user_data[user]['chatId'],'There is a new version of EazeBot available on git/pip (v%s) with these changes:\n%s\n\nConsider updating!'%(remoteVersion,versionMessage)))

def updateTradeSets(bot,job):
    updater = job.context
    logging.info('Updating trade sets...')
    for user in updater.dispatcher.user_data:
        if user in __config__['telegramUserId'] and 'trade' in updater.dispatcher.user_data[user]:
            for iex,ex in enumerate(updater.dispatcher.user_data[user]['trade']):
                try: # make sure other exchanges are checked too, even if one has a problem
                    updater.dispatcher.user_data[user]['trade'][ex].update()
                except Exception as e:
                    logging.error(str(e))
                    pass
    logging.info('Finished updating trade sets...')

def updateBalance(bot,job):
    updater = job.context
    logging.info('Updating balances...')
    for user in updater.dispatcher.user_data:
        if user in __config__['telegramUserId'] and 'trade' in updater.dispatcher.user_data[user]:
            for iex,ex in enumerate(updater.dispatcher.user_data[user]['trade']):
                updater.dispatcher.user_data[user]['trade'][ex].updateBalance()
    logging.info('Finished updating balances...')
    
def checkCandle(bot,job,which=1):
    updater = job.context
    logging.info('Checking candles for all trade sets...')
    for user in updater.dispatcher.user_data:
        if user in __config__['telegramUserId']:
            for iex,ex in enumerate(updater.dispatcher.user_data[user]['trade']):
                # avoid to hit it during updating
                updater.dispatcher.user_data[user]['trade'][ex].update(specialCheck=which)
    logging.info('Finished checking candles for all trade sets...')

def timingCallback(bot, update,user_data,query=None,response=None):
    if query is None:
        query = update.callback_query
    if query is None:
        return 0
    query.message.delete()
    if 'Yes' in query.data:
        query.answer('Please give the price above which the daily candle should close in order to initiate the buy!')
        return NUMBER
    else:
        query.answer()
        return user_data['lastFct'].pop()(None)
   
def showSettings  (bot, update,user_data,botOrQuery=None):
    # show gain/loss in fiat
    # give preferred fiat
    # stop bot with security question
    string = '*Settings:*\n_Fiat currencies(descending priority):_ %s\n_Show gain/loss in:_ %s'%(', '.join(user_data['settings']['fiat']), 'Fiat (if available)' if user_data['settings']['showProfitIn'] is not None else 'Base currency') #user_data['settings']['showProfitIn']
    settingButtons = [[InlineKeyboardButton('Define your fiat',callback_data='settings|defFiat')],[InlineKeyboardButton("Toggle showing gain/loss in baseCurrency or fiat", callback_data='settings|toggleProfit')],[InlineKeyboardButton("*Stop bot*", callback_data='settings|stopBot'),InlineKeyboardButton("Back", callback_data='settings|cancel')]]    
    if botOrQuery == None or isinstance(botOrQuery,type(bot)):
        user_data['messages']['settings'].append(bot.send_message(user_data['chatId'], string, parse_mode = 'markdown', reply_markup=InlineKeyboardMarkup(settingButtons)))
    else:
        try:
            botOrQuery.answer('Settings updated')
            botOrQuery.edit_message_text(string, parse_mode = 'markdown', reply_markup=InlineKeyboardMarkup(settingButtons))
        except BadRequest:
            user_data['messages']['settings'].append(bot.send_message(user_data['chatId'], string, parse_mode = 'markdown', reply_markup=InlineKeyboardMarkup(settingButtons)))
        
def updateTStext(bot,update,user_data,uidTS,query=None):
    if query:
        try:
            query.message.delete()
        except:
            pass
    printTradeStatus(bot,update,user_data,uidTS)

    
def InlineButtonCallback(bot, update,user_data,query=None,response=None):
    if query is None:
        query = update.callback_query
    if query is None:
        return 0
    command,*args = query.data.split('|')
    if 'chosen' in args:
        tmp = query.data.split('|')
        tmp.remove('chosen')
        query.data = '|'.join(tmp)
    
    if 'cancel' in args:
        user_data['tempTradeSet'] = [None,None,None]
        deleteMessages(user_data,['dialog','botInfo','settings'])
    else:
        if command == 'settings':
            subcommand = args.pop(0)
            if subcommand == 'stopBot':
                if len(args) == 0:
                    query.answer('')
                    bot.send_message(user_data['chatId'],'Are you sure you want to stop the bot? *Caution! You have to restart the Python script; until then the bot will not be responding to Telegram input!*',parse_mode = 'markdown', reply_markup = InlineKeyboardMarkup( [[InlineKeyboardButton('Yes',callback_data='settings|stopBot|Yes')],[InlineKeyboardButton("No", callback_data='settings|cancel')]]) )
                elif args[0] == 'Yes':
                    query.answer('stopping')
                    bot.send_message(user_data['chatId'],'Bot is aborting now. Goodbye!')
                    doneCmd(bot,update,user_data)
            else:
                if subcommand == 'defFiat':
                    if response is None:
                        user_data['lastFct'].append(lambda res : InlineButtonCallback(bot,update,user_data,query,res))
                        user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'Please name your fiat currencies (e.g. USD). You can also name multiple currencies separated with commata,  \
                                         (e.g. type: USD,USDT,TUSD) such that in case the first currency does not exist on an exchange, the second one is used, and so on.'))
                        return INFO
                    else:
                        user_data['settings']['fiat'] =  response.upper().split(',')
                        
                elif subcommand == 'toggleProfit':
                        if user_data['settings']['showProfitIn'] is None:
                            if len(user_data['settings']['fiat'])>0:
                                user_data['settings']['showProfitIn'] = user_data['settings']['fiat']
                            else:
                                query.answer('Please first specify fiat currency(s) in the settings.')
                        else:
                            user_data['settings']['showProfitIn'] = None
                showSettings(bot, update,user_data,query)
        else:
            exch = args.pop(0)
            uidTS = args.pop(0)
            if command == 'toggleCurrency':
                user_data['whichCurrency'] = (user_data['whichCurrency']+1)%2
                return askAmount(user_data,exch,uidTS,args[0],query)
            
            elif command == 'showSymbols':
                syms = [val for val in user_data['trade'][exch].exchange.symbols if not '.d' in val]
                buttons = list()
                rowbuttons = []
                string = ''                        
                for count, sym in enumerate(syms):
                    if count % 4 == 0: # 4 buttons per row
                        if len(rowbuttons) > 0:
                            buttons.append(rowbuttons)
                        rowbuttons = [InlineKeyboardButton(sym,callback_data='chooseSymbol|%s|%s'%(exch,sym))]
                    else:
                        rowbuttons.append(InlineKeyboardButton(sym,callback_data='chooseSymbol|%s|%s'%(exch,sym)))
                    string += (sym + ', ')
                buttons.append(rowbuttons)
                buttons.append([InlineKeyboardButton('Cancel',callback_data='xxx|cancel')])
                try:
                    query.edit_message_text('Choose a pair...',reply_markup=InlineKeyboardMarkup(buttons))
                except BadRequest as e:
                    query.edit_message_text('Too many pairs to make buttons, you have to type the pair. Here is a list of all pairs:\n'+string[0:-2],reply_markup=[])
                
            elif command == 'chooseSymbol' :
                query.message.delete()
                return user_data['lastFct'].pop()(uidTS)  # it is no uidTS but the chosen symbol..i was too lazy to use new variable ;-)
            
            elif command == '1':   # donations
                if len(args) > 0:
                    if exch == 'xxx':
                        # get all exchange names that list the chosen coin and ask user from where to withdraw
                        exchs = [ct.exchange.name for _,ct in user_data['trade'].items() if args[0] in ct.exchange.currencies]
                        buttons = [[InlineKeyboardButton(exch,callback_data='1|%s|%s|%s'%(exch.lower(),'xxx',args[0]))] for exch in sorted(exchs)] + [[InlineKeyboardButton('Cancel',callback_data='1|xxx|xxx|cancel')]]
                        query.edit_message_text('From which exchange listing %s do you want to donate?'%args[0],reply_markup=InlineKeyboardMarkup(buttons))
                        query.answer('')
                    else:
                        if response is not None:
                            if args[0] == 'BTC':
                                address = '3AP2u8wMwdSFJWCXNhUbfbV1xirqshfqg6'
                            elif args[0] == 'ETH':
                                address = '0xE0451300D96090c1F274708Bc00d791017D7a5F3'
                            elif args[0] == 'NEO':
                                address = 'AaGRMPuwtGrudXR5s7F5n11cxK595hCWUg' 
                            elif args[0] == 'XLM':
                                address = 'GBJEFEFUAUVTWL5UYK3NTWW7J5J3SMH4XB7SYDZRWWEON5S5YHPI2LAR'
                            try:
                                if response > 0:
                                    user_data['trade'][exch].exchange.withdraw(args[0], response, address)
                                    bot.send_message(user_data['chatId'],'Donation suceeded, thank you very much!!!')
                                else:
                                    bot.send_message(user_data['chatId'],'Amount <= 0 %s. Donation canceled =('%args[0])
                            except Exception as e:
                                bot.send_message(user_data['chatId'],'There was an error during withdrawing, thus donation failed! =( Please consider the following reasons:\n- Insufficient funds?\n-2FA authentication required?\n-API key has no withdrawing permission?\n\nServer response was:\n<i>%s</i>'%str(e),parse_mode='html')
                        else:
                            ct = user_data['trade'][exch]
                            balance = ct.exchange.fetch_balance()
                            if ct.exchange.fees['funding']['percentage']:
                                query.answer('')
                                bot.send_message(user_data['chatId'],'Error. Exchange using relative withdrawal fees. Not implemented, please contact developer.')
                            if balance['free'][args[0]] > ct.exchange.fees['funding']['withdraw'][args[0]]:
                                query.answer('')
                                bot.send_message(user_data['chatId'],'Your free balance is %.8g %s and withdrawing fee on %s is %.8g %s. How much do you want to donate (excluding fees)'%(balance['free'][args[0]],args[0],exch,ct.exchange.fees['funding']['withdraw'][args[0]],args[0])) 
                                user_data['lastFct'].append(lambda res : InlineButtonCallback(bot,update,user_data,query,res))
                                return NUMBER
                            else:
                                query.answer('%s has insufficient free %s. Choose another exchange!'%(exch,args[0])) 
                else:
                    buttons = [[InlineKeyboardButton("Donate BTC",callback_data='1|%s|%s|BTC'%('xxx','xxx')),InlineKeyboardButton("Donate ETH",callback_data='%s|%s|%d|ETH'%('xxx','xxx',1)),InlineKeyboardButton("Donate NEO",callback_data='1|%s|%s|NEO'%('xxx','xxx')),InlineKeyboardButton("Donate XLM",callback_data='1|%s|%s|XLM'%('xxx','xxx'))]] 
                    query.edit_message_text('Thank you very much for your intention to donate some crypto! Accepted coins are BTC, ETH and NEO.\nYou may either donate by sending coins manually to one of the addresses below, or more easily by letting the bot send coins (amount will be asked in a later step) from one of your exchanges by clicking the corresponding button below.\n\n*BTC address:*\n17SfuTsJ3xpbzgArgRrjYSjvmzegMRcU3L\n*ETH address:*\n0x2DdbDA69B27D36D0900970BCb8049546a9d621Ef\n*NEO address:*\nAaGRMPuwtGrudXR5s7F5n11cxK595hCWUg'  ,reply_markup=InlineKeyboardMarkup(buttons),parse_mode='markdown')
            
            elif command == 'chooseExch':
                query.answer('%s chosen'%exch)
                query.message.delete()
                return user_data['lastFct'].pop()(exch)
            
            elif command == 'resetTSH':
                ct = user_data['trade'][exch]
                if 'yes' in args:
                    query.message.delete()
                    ct.resetTradeHistory()
                elif 'no' in args:
                    query.edit_message_reply_markup(reply_markup=buttonsEditTSH(ct))
                else:
                    query.answer('Are you sure? This cannot be undone!')
                    query.edit_message_reply_markup(reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton("Yes", callback_data='resetTSH|%s|XXX|yes'%(exch)), InlineKeyboardButton("No", callback_data='resetTSH|%s|XXX|no'%(exch))]]))
                    
            else:  # trade set commands
                if exch not in user_data['trade'] or uidTS not in user_data['trade'][exch].tradeSets:
                    query.edit_message_reply_markup()
                    query.edit_message_text('This trade set is not found anymore. Probably it was deleted')
                else:
                    ct = user_data['trade'][exch]                
                    if command == '2':  # edit trade set
                        if 'chosen' in args:
                            try:
                                query.edit_message_reply_markup(reply_markup=makeTSInlineKeyboard(exch,uidTS))
                            except:
                                pass
                            
                        if 'back' in args:
                            query.answer('')
                            
                        elif any(['BLD' in val for val in args]):    
                            ct.deleteBuyLevel(uidTS,int([re.search('(?<=^BLD)\d+',val).group(0) for val in args if isinstance(val,str) and 'BLD' in val][0]))
                            updateTStext(bot,update,user_data,uidTS,query)
                            query.answer('Deleted buy level')
                        
                        elif any(['SLD' in val for val in args]):    
                            ct.deleteSellLevel(uidTS,int([re.search('(?<=^SLD)\d+',val).group(0) for val in args if isinstance(val,str) and 'SLD' in val][0]))
                            updateTStext(bot,update,user_data,uidTS,query)
                            query.answer('Deleted sell level')
                        
                        elif 'buyAdd' in args:   
                            if response is None:
                                query.answer('Adding new buy level')
                                return askPos(bot,user_data,exch,uidTS,direction='buy',applyFct=lambda : InlineButtonCallback(bot,update,user_data,query,'continue'))
                            else:
                                ct.addBuyLevel(uidTS,user_data['tempTradeSet'][0],user_data['tempTradeSet'][1],user_data['tempTradeSet'][2])
                                user_data['tempTradeSet'] = [None,None,None]
                                updateTStext(bot,update,user_data,uidTS,query)
                        
                        elif 'sellAdd' in args:
                            if response is None:
                                query.answer('Adding new sell level')
                                return askPos(bot,user_data,exch,uidTS,direction='sell',applyFct=lambda : InlineButtonCallback(bot,update,user_data,query,'continue'))
                            else:
                                ct.addSellLevel(uidTS,user_data['tempTradeSet'][0],user_data['tempTradeSet'][1])
                                user_data['tempTradeSet'] = [None,None,None]
                                updateTStext(bot,update,user_data,uidTS,query)
                                
                        elif any(['buyReAdd' in val for val in args]):
                            logging.info(args)
                            level = int([re.search('(?<=^buyReAdd)\d+',val).group(0) for val in args if isinstance(val,str) and 'buyReAdd' in val][0])
                            trade = ct.tradeSets[uidTS]['InTrades'][level]
                            ct.addBuyLevel(uidTS,trade['price'],trade['amount'],trade['candleAbove'])                             
                            updateTStext(bot,update,user_data,uidTS,query)
                        
                        elif any(['sellReAdd' in val for val in args]):
                            level = int([re.search('(?<=^sellReAdd)\d+',val).group(0) for val in args if isinstance(val,str) and 'sellReAdd' in val][0])
                            trade = ct.tradeSets[uidTS]['OutTrades'][level]
                            ct.addSellLevel(uidTS,trade['price'],trade['amount'])                             
                            updateTStext(bot,update,user_data,uidTS,query)
                            
                        elif 'AIC' in args:
                            query.answer('Adding initial coins')
                            return addInitBalance(bot,user_data,exch,uidTS,inputType=None,response=None,fct = lambda :printTradeStatus(bot,update,user_data,uidTS))
                        
                        elif 'TSgo' in args:
                            ct.activateTradeSet(uidTS)
                            updateTStext(bot,update,user_data,uidTS,query)
                            query.answer('Trade set activated')
                            
                        elif 'TSstop' in args:
                            ct.deactivateTradeSet(uidTS,1)
                            updateTStext(bot,update,user_data,uidTS,query)
                            query.answer('Trade set deactivated!')
                            
                        elif 'SLM' in args:
                            buttons = buttonsSL(ct,uidTS)
                            query.edit_message_reply_markup(reply_markup=InlineKeyboardMarkup(buttons))
                            query.answer('Choose an option')                          
                            
                        elif 'SLBE' in args:
                            ans = ct.setSLBreakEven(uidTS)
                            if ans:
                                query.answer('SL set break even')
                            else:
                                query.answer('SL break even failed to set')
                            updateTStext(bot,update,user_data,uidTS,query)
                        
                        elif 'DCSL' in args:
                            # set a daily-close SL
                            if response is None:
                                if 'yes' in args:
                                    query.message.delete()
                                    user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'Please give the daily candle closing price below which a SL would be triggered'))
                                    return NUMBER
                                else:
                                    user_data['lastFct'].append(lambda res : InlineButtonCallback(bot,update,user_data,query,res))
                                    user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'Do you want to set an SL, which is triggered when the daily candle closes below price X?',reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton("Yes", callback_data='2|%s|%s|DCSL|yes'%(exch,uidTS)),InlineKeyboardButton("No", callback_data='2|%s|%s|cancel'%(exch,uidTS))], [InlineKeyboardButton("Cancel", callback_data='2|%s|%s|cancel'%(exch,uidTS))]])))
                            else:
                                ct.setDailyCloseSL(uidTS,response)
                                deleteMessages(user_data,'dialog')
                                updateTStext(bot,update,user_data,uidTS,query)
                                
                        elif 'WCSL' in args:
                            # set a daily-close SL
                            if response is None:
                                if 'yes' in args:
                                    query.message.delete()
                                    user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'Please give the weekly candle closing price below which a SL would be triggered'))
                                    return NUMBER
                                else:
                                    user_data['lastFct'].append(lambda res : InlineButtonCallback(bot,update,user_data,query,res))
                                    user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'Do you want to set an SL, which is triggered when the weekly candle closes below price X?',reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton("Yes", callback_data='2|%s|%s|WCSL|yes'%(exch,uidTS)),InlineKeyboardButton("No", callback_data='2|%s|%s|cancel'%(exch,uidTS))], [InlineKeyboardButton("Cancel", callback_data='2|%s|%s|cancel'%(exch,uidTS))]])))
                            else:
                                ct.setWeeklyCloseSL(uidTS,response)
                                deleteMessages(user_data,'dialog')
                                updateTStext(bot,update,user_data,uidTS,query)
                            
                        elif 'TSL' in args:
                            # set a trailing stop-loss
                            if response is None:
                                query.answer()
                                if 'abs' in args or 'rel' in args:
                                    query.message.delete()
                                    user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],'Please enter the trailing SL offset%s'%('' if 'abs' in args else ' in %')))
                                    user_data['lastFct'].append(lambda res : InlineButtonCallback(bot,update,user_data,query,res))
                                    return NUMBER
                                else:
                                    user_data['messages']['dialog'].append(bot.send_message(user_data['chatId'],"What kind of trailing stop-loss offset do you want to set?", reply_markup = InlineKeyboardMarkup([[InlineKeyboardButton("Absolute", callback_data='2|%s|%s|TSL|abs'%(exch,uidTS)),InlineKeyboardButton("Relative", callback_data='2|%s|%s|TSL|rel'%(exch,uidTS))]])))
                            else:
                                response = float(response)
                                if 'rel' in args:
                                    response /= 100
                                ct.setTrailingSL(uidTS,response,typ='abs' if 'abs' in args else 'rel')
                                deleteMessages(user_data,'dialog')
                                updateTStext(bot,update,user_data,uidTS,query)
                                
                        elif 'SLC' in args:
                            if response is None:
                                query.answer('Please enter the new SL (0 = no SL)')
                                user_data['lastFct'].append(lambda res : InlineButtonCallback(bot,update,user_data,query,res))
                                return NUMBER
                            else:
                                response = float(response)
                                if response == 0:
                                    response = None
                                ct.setSL(uidTS,response)
                                updateTStext(bot,update,user_data,uidTS,query)
                        else: 
                            buttons = buttonsEditTS(ct,uidTS,'full')
                            query.edit_message_reply_markup(reply_markup=InlineKeyboardMarkup(buttons))
                            query.answer('Choose an option')
                    elif command == '3':  # init trade set deletion
                        if 'ok' in args:
                            query.message.delete()
                            ct.deleteTradeSet(uidTS,sellAll='yes' in args)
                            query.answer('Trade Set deleted')
                        elif 'yes' in args or 'no' in args:
                            query.answer('Ok, and are you really sure to delete this trade set?')
                            query.edit_message_reply_markup(reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton("Yes", callback_data='3|%s|%s|ok|%s'%(exch,uidTS,'|'.join(args))),InlineKeyboardButton("Cancel",callback_data='3|%s|%s|cancel'%(exch,uidTS))]]))
                        else:
                            query.answer('Do you want to sell your remaining coins?')
                            query.edit_message_reply_markup(reply_markup=InlineKeyboardMarkup([[InlineKeyboardButton("Yes", callback_data='3|%s|%s|yes'%(exch,uidTS)),InlineKeyboardButton("No", callback_data='3|%s|%s|no'%(exch,uidTS))]]))
    return MAINMENU

def clean_data(user_data):
    delThese = []
    for user in user_data:
        # discard unknown users
        if not (user in __config__['telegramUserId'] and 'trade' in user_data[user]):
            delThese.append(user)
        else: # discard cached messages
            if 'messages' in user_data[user]:
                deleteMessages(user_data[user],'all',True)
    for k in delThese:
        user_data.pop(k, None)
    return user_data

def save_data(*arg):
    if len(arg) == 1:
        updater = arg[0]
    else:
        bot,job = arg
        updater = job.context
    # remove backup
    try:
        os.remove('data.bkp') 
    except:
        logging.warning('No backup file found')
        pass
    # rename last save to backup
    try:
        os.rename('data.pickle', 'data.bkp')
        logging.info('Created user data autosave backup')
    except:
        logging.warning('Could not rename last saved data to backup')
        pass
    user_data = deepcopy(updater.dispatcher.user_data)
    clean_data(user_data)
    # write user data
    with open('data.pickle', 'wb') as f:
        dill.dump(user_data, f)
    logging.info('User data autosaved')
        
def convert_data(from_='linux',to_='win',filename='data.pickle',filenameout='data.pickle.new'):
    with open(filename, 'rb') as fi:
        byteContent = fi.read()
        with open(filenameout, 'wb') as fo:
            if 'linux' in from_ and 'win' in to_:
                byteContent = byteContent.replace(b'cdill._dill', b'cdill.dill')
            elif 'win' in from_ and 'linux' in to_:
                byteContent = byteContent.replace(b'cdill.dill',b'cdill._dill')
            if 'git' not in from_ and 'git' in to_:
                byteContent = byteContent.replace(b'ceazebot.tradeHandler', b'ctradeHandler')
                byteContent = byteContent.replace(b'ceazebot.EazeBot', b'cEazeBot')
            elif 'git' in from_ and 'git' not in to_:
                byteContent = byteContent.replace(b'ctradeHandler', b'ceazebot.tradeHandler')
                byteContent = byteContent.replace(b'cEazeBot', b'ceazebot.EazeBot')
            fo.write(byteContent)
        
        
def load_data(filename='data.pickle'):
    # load latest user data
    if os.path.isfile(filename):
        try:
            with open(filename, 'rb') as f:
                logging.info('Loading user data')
                return dill.load(f)
        except Exception as e:
            raise(e)
    else:
        logging.error('No autosave file found')
        return defaultdict(dict)    
    
def startBot():
    print('\n\n******** Welcome to EazeBot (v%s) ********\nFree python/telegram bot for easy execution and surveillance of crypto trading plans on multiple exchanges\n\n'%thisVersion)
    global __config__
    global job_queue
    global updater
    #%% load bot configuration
    with open("botConfig.json", "r") as fin:
        __config__ = json.load(fin)
    if isinstance(__config__['telegramUserId'],str) or isinstance(__config__['telegramUserId'],int):
        __config__['telegramUserId'] = [int(__config__['telegramUserId'])]
    elif isinstance(__config__['telegramUserId'],list):
        __config__['telegramUserId'] = [int(val) for val in __config__['telegramUserId']]
    if isinstance(__config__['updateInterval'],str):
        __config__['updateInterval'] = int(__config__['updateInterval'])
    if 'minBalanceInBTC' not in __config__:
        __config__['minBalanceInBTC'] = 0.001
    if isinstance(__config__['minBalanceInBTC'],str):
        __config__['minBalanceInBTC'] = float(__config__['minBalanceInBTC'])
    if 'debug' not in __config__:
        __config__['debug'] = False
    if isinstance(__config__['debug'],str):
        __config__['debug'] = bool(int(__config__['debug']))
    
    #%% define the handlers to communicate with user
    conv_handler = ConversationHandler(
            entry_points=[CommandHandler('start', startCmd,pass_user_data=True)],
            states={
                MAINMENU: [RegexHandler('^Status of Trade Sets$',printTradeStatus,pass_user_data=True),
                            RegexHandler('^New Trade Set$',createTradeSet,pass_user_data=True),
                            RegexHandler('^Trade History$',printTradeHistory,pass_user_data=True),
                            RegexHandler('^Add/update exchanges',addExchanges,pass_user_data=True),
                            RegexHandler('^Bot Info$',botInfo,pass_user_data=True),
                            RegexHandler('^Check Balance$',checkBalance,pass_user_data=True),
                            RegexHandler('^Settings$',showSettings,pass_user_data=True),
                            CallbackQueryHandler(InlineButtonCallback,pass_user_data=True),
                            MessageHandler(Filters.text,lambda b,u: noncomprende(b,u,'noValueRequested'))],
                SYMBOL:   [RegexHandler('\w+/\w+',receivedInfo,pass_user_data=True),
                            MessageHandler(Filters.text,lambda b,u: noncomprende(b,u,'wrongSymbolFormat')),
                            CallbackQueryHandler(InlineButtonCallback,pass_user_data=True)],
                NUMBER:   [RegexHandler('^[\+,\-]?\d+\.?\d*$',receivedFloat,pass_user_data=True),
                           MessageHandler(Filters.text,lambda b,u: noncomprende(b,u,'noNumber')),
                           CallbackQueryHandler(InlineButtonCallback,pass_user_data=True)],
                TIMING:   [CallbackQueryHandler(timingCallback,pass_user_data=True),
                           CallbackQueryHandler(InlineButtonCallback,pass_user_data=True)],
                INFO:     [RegexHandler('\w+',receivedInfo,pass_user_data=True)]
            },
            fallbacks=[CommandHandler('exit', doneCmd,pass_user_data=True)], allow_reentry = True)#, per_message = True)
    unknown_handler = MessageHandler(Filters.command, lambda b,u: noncomprende(b,u,'unknownCmd'))
    
    #%% start telegram API, add handlers to dispatcher and start bot
    updater = Updater(token = __config__['telegramAPI'], request_kwargs={'read_timeout': 10, 'connect_timeout': 10})
    job_queue = updater.job_queue
    updater.dispatcher.add_handler(conv_handler)
    updater.dispatcher.add_handler(unknown_handler)
    updater.dispatcher.user_data = clean_data(load_data())
    
    for user in __config__['telegramUserId']:
        if user in updater.dispatcher.user_data and len(updater.dispatcher.user_data[user]) > 0:
            time.sleep(2) # wait because of possibility of temporary exchange lockout
            addExchanges(updater.bot,None,updater.dispatcher.user_data[user])
    
    for user in __config__['telegramUserId']:
        try:
            updater.bot.send_message(user,'Bot was restarted.\n Please press /start to continue.',reply_markup=ReplyKeyboardMarkup([['/start']]),one_time_keyboard=True)
        except:
            pass
    
    # start a job updating the trade sets each interval
    updater.job_queue.run_repeating(updateTradeSets, interval=60*__config__['updateInterval'], first=60,context=updater)
    # start a job checking for updates once a  day
    updater.job_queue.run_repeating(checkForUpdates, interval=60*60*24, first=0,context=updater)
    # start a job checking every day 10 sec after midnight (UTC time)
    updater.job_queue.run_daily(lambda u,b: checkCandle(u,b,1), (dt.datetime.combine(dt.date(1900,5,5),dt.time(0,0,10)) + (dt.datetime.now()-dt.datetime.utcnow())).time(), context=updater,name='dailyCheck')
    # start a job checking every week 10 sec after midnight (UTC time)
    updater.job_queue.run_daily(lambda u,b: checkCandle(u,b,2), (dt.datetime.combine(dt.date(1900,5,5),dt.time(0,0,10)) + (dt.datetime.now()-dt.datetime.utcnow())).time(), days=tuple([0]),context=updater,name='weeklyCheck')
    # start a job saving the user data each 5 minutes
    updater.job_queue.run_repeating(save_data, interval=5*60, first=75,context=updater)
    updater.start_polling()
    updater.idle()
    save_data(updater)  # last data save when finishing

# execute main if running as script
if __name__ == '__main__':
    startBot()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# EazeBot
# Free python/telegram bot for easy execution and surveillance of crypto trading plans on multiple exchanges.
# Copyright (C) 2018
# Marcel Beining <marcel.beining@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This class is used to control trading sets"""

import ccxt
import re
from json import JSONDecodeError
from inspect import getsourcefile, getsourcelines
import numpy as np
import time
import datetime
import random
import string
import sys, os
from ccxt.base.errors import (AuthenticationError,NetworkError,OrderNotFound,InvalidNonce,ExchangeError)

class tradeHandler:
    
    def __init__(self,exchName,key=None,secret=None,password=None,uid=None,messagerFct=None,logger=None):
        # use either the given messager function or define a simple print messager function which takes a level argument as second optional input
        if messagerFct:
            self.message = messagerFct
        else:
            self.message = lambda a,b='Info': print(b + ': ' + a)
        
        self.logger = logger
        checkThese = ['cancelOrder','createLimitOrder','fetchBalance','fetchTicker']
        self.tradeSets = {}
        self.tradeSetHistory = []
        self.exchange = getattr (ccxt, exchName) ({'enableRateLimit': True,'options': { 'adjustForTimeDifference': True }}) # 'nonce': ccxt.Exchange.milliseconds,
        if key:
            self.exchange.apiKey = key
        if secret:
            self.exchange.secret = secret
        if password:
            self.exchange.password = password
        if uid:
            self.exchange.uid = uid

        self.updating = False
        self.waiting = []
        self.down = False
        self.authenticated = False
        if key:
            self.updateKeys(key,secret,password,uid)
                        
        if not all([self.exchange.has[x] for x in checkThese]):
            text = 'Exchange %s does not support all required features (%s)'%(exchName,', '.join(checkThese))
            self.message(text,'error')
            raise Exception(text)
        self.amount2Prec = lambda a,b: self.stripZeros(self.exchange.amountToPrecision(a,b)) if isinstance(self.exchange.amountToPrecision(a,b),str) else self.stripZeros(format(self.exchange.amountToPrecision(a,b),'.10f'))
        self.price2Prec = lambda a,b: self.stripZeros(self.exchange.priceToPrecision(a,b)) if isinstance(self.exchange.priceToPrecision(a,b),str) else self.stripZeros(format(self.exchange.priceToPrecision(a,b),'.10f'))
        self.cost2Prec = lambda a,b: self.stripZeros(self.exchange.costToPrecision(a,b)) if isinstance(self.exchange.costToPrecision(a,b),str) else self.stripZeros(format(self.exchange.costToPrecision(a,b),'.10f'))
        self.fee2Prec = lambda a,b: self.stripZeros(str(b))
        
        
          
    def __reduce__(self):
        # function needes for serializing the object
        return (self.__class__, (self.exchange.__class__.__name__,None,None,None,None,self.message),self.__getstate__(),None,None)
    
    def __setstate__(self,state):
        if isinstance(state,tuple):
            state,tshs = state
        else:
            tshs = []
        for iTs in state: # temp fix for old trade sets that do not some of the newer fields
            ts = state[iTs]
            # clear updating variables from old state
            ts['waiting'] = []
            ts['updating'] = False
            if 'trailingSL' not in ts:
                ts['trailingSL'] = [None,None]
            if 'dailycloseSL' not in ts:
                ts['dailycloseSL'] = None  
            if 'weeklycloseSL' not in ts:
                ts['weeklycloseSL'] = None 
            for trade in ts['InTrades']:
                if 'actualAmount' not in trade:
                    fee = self.exchange.calculateFee(ts['symbol'],'limit','buy',trade['amount'],trade['price'],'maker')
                    if fee['currency'] == ts['coinCurrency']:
                        trade['actualAmount'] = trade['amount'] - (fee['cost'] if self.exchange.name.lower() != 'binance' or self.getFreeBalance('BNB') < 0.5 else 0) # this is a hack, as fees on binance are deduced from BNB if this is activated and there is enough BNB, however so far no API chance to see if this is the case. Here I assume that 0.5 BNB are enough to pay the fee for the trade and thus the fee is not subtracted from the traded coin
                    else:
                        trade['actualAmount'] = trade['amount']
        self.tradeSets = state
        self.tradeSetHistory = tshs
        
    def __getstate__(self):
        if hasattr(self,'tradeSetHistory'):
            return (self.tradeSets,self.tradeSetHistory)
        else:
            return (self.tradeSets,[])
    
    @staticmethod
    def stripZeros(string):
        if '.' in string:
            return string.rstrip('0').rstrip('.')
        else:
            return string
        
    def checkNum(self,*value):
        return all([(isinstance(val,float) | isinstance(val,int)) if not isinstance(val,list) else self.checkNum(*val) for val in value])
    
    def checkQuantity(self,symbol,typ,qty):
        if typ not in ['amount','price','cost']:
            raise ValueError('Type is not amount, price or cost')
        if typ in self.exchange.markets[symbol]['limits']:
            return (self.exchange.markets[symbol]['limits'][typ]['min'] is None or qty >= self.exchange.markets[symbol]['limits'][typ]['min']) and (self.exchange.markets[symbol]['limits'][typ]['max'] is None or self.exchange.markets[symbol]['limits'][typ]['max'] == 0 or qty <= self.exchange.markets[symbol]['limits'][typ]['max'])
        else:
            if self.logger:
                self.logger.warning('Exchange %s does not provide limits for %s'%(self.exchange.name,typ))
            return True
        
    def safeRun(self,func,printError=True,iTs=None):
        count = 0
        wasdown = self.down
        while True:
            try:
                self.down = False
                return func()
            except NetworkError as e:
                count += 1
                if hasattr(self.exchange, 'load_time_difference'):
                    self.exchange.load_time_difference()
                if count >= 5:
                    if iTs:
                        self.unlockTradeSet(iTs)
                    if 'Cloudflare' in str(e):
                        self.down = True
                        if printError:
                             self.message('Cloudflare problem with exchange %s. Exchange is treated as down.'%(self.exchange.name,'' if iTs is None else ' for tradeSet %d (%s)'%(list(self.tradeSets.keys()).index(iTs))),'Error')   
                    elif printError:
                        self.message('Network exception occurred 5 times in a row on %s%s'%(self.exchange.name,'' if iTs is None else ' for tradeSet %d (%s)'%(list(self.tradeSets.keys()).index(iTs),self.tradeSets[iTs]['symbol'])),'Error')
                    raise(e)
                else:
                    time.sleep(0.5)
                    continue
            except OrderNotFound as e:
                count += 1
                if count >= 5:
                    if iTs:
                        self.unlockTradeSet(iTs)
                    if printError:
                        self.message('Order not found error 5 times in a row on %s%s'%(self.exchange.name,'' if iTs is None else ' for tradeSet %d (%s)'%(list(self.tradeSets.keys()).index(iTs),self.tradeSets[iTs]['symbol'])),'Error')
                    raise(e)
                else:
                    time.sleep(0.5)
                    continue
            except AuthenticationError as e:
                count += 1
                if count >= 5:
                    if iTs:
                        self.unlockTradeSet(iTs)
                    raise(e)
                else:
                    time.sleep(0.5)
                    continue
            except InvalidNonce as e:
                count += 1
                # this tries to resync the system timestamp with the exchange's timestamp
                if hasattr(self.exchange, 'load_time_difference'):
                    self.exchange.load_time_difference()
                time.sleep(0.5)
                continue
            except JSONDecodeError as e:
                if iTs:
                    self.unlockTradeSet(iTs)
                if 'Expecting value' in str(e):
                    self.down = True
                    if printError:
                        self.message('%s seems to be down.'%self.exchange.name)   
                raise(e)
            except Exception as e:
                if count < 4 and isinstance(e,ExchangeError) and "symbol" in str(e).lower():
                    self.safeRun(self.exchange.loadMarkets)
                    count += 1
                    continue
                elif count < 4 and ('unknown error' in str(e).lower() or 'connection' in str(e).lower()):
                    count += 1
                    time.sleep(0.5)
                    continue
                else:
                    if iTs:
                        self.unlockTradeSet(iTs)
                    string = 'Exchange %s\n'%self.exchange.name
                    if count >= 5:
                        string += 'Network exception occurred 5 times in a row! Last error was:\n'
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    #fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                    lines = getsourcelines(func)
                    string += '%s in %s from %s at line %d: %s'%(exc_type, lines[0][0],os.path.basename(getsourcefile(func)),lines[1],str(e))

                    if printError:
                        self.message(string,'Error')
                    raise(e)
            finally:
                if wasdown and not self.down:
                    self.message('Exchange %s seems back to work!'%self.exchange.name)
        

    def lockTradeSet(self,iTs):
        # avoids two processes changing a tradeset at the same time
        count = 0
        mystamp = time.time()
        self.tradeSets[iTs]['waiting'].append(mystamp)
        time.sleep(0.2)
        while self.tradeSets[iTs]['updating'] or self.tradeSets[iTs]['waiting'][0] < mystamp:
            count += 1
            time.sleep(1)		
            if count > 60: # 60 sec max wait
                try:  # cautionary so that no timestamp can stay in the queue due to some messaging error
                    self.message('Waiting for tradeSet update (%s on %s) to finish timed out after 1 min.. Resetting updating variable now.'%(self.tradeSets[iTs]['symbol'],self.exchange.name),'error')
                except:
                    pass
                break
        self.tradeSets[iTs]['updating'] = True
        self.tradeSets[iTs]['waiting'].remove(mystamp)
    
    def unlockTradeSet(self,iTs):
        self.tradeSets[iTs]['updating'] = False
        
    def updateBalance(self):
        self.updateDownState(True)
        # reloads the exchange market and private balance and, if successul, sets the exchange as authenticated
        self.safeRun(self.exchange.loadMarkets) 
        self.balance = self.safeRun(self.exchange.fetch_balance)
        self.authenticated = True
        
    def getFreeBalance(self,coin):
        if coin in self.balance:
            return self.balance[coin]['free']
        else:
            return 0

    def updateKeys(self,key=None,secret=None,password=None,uid=None):
        if key:
            self.exchange.apiKey = key
        if secret:
            self.exchange.secret = secret
        if password:
            self.exchange.password = password
        if uid:
            self.exchange.uid = uid
        try: # check if keys work
            self.updateBalance()
        except AuthenticationError as e:#
            self.authenticated = False
            try:
                self.message('Failed to authenticate at exchange %s. Please check your keys'%self.exchange.name,'error')
            except:
                print('Failed to authenticate at exchange %s. Please check your keys'%self.exchange.name)
        except getattr(ccxt,'ExchangeError') as e:
            self.authenticated = False
            if 'key' in str(e).lower():
                try:
                    self.message('Failed to authenticate at exchange %s. Please check your keys'%self.exchange.name,'error')
                except:
                    print('Failed to authenticate at exchange %s. Please check your keys'%self.exchange.name)
            else:
                try:
                    self.message('The following error occured at exchange %s:\n%s'%(self.exchange.name,str(e)),'error')
                except:
                    print('The following error occured at exchange %s:\n%s'%(self.exchange.name,str(e)))
          
    
    def initTradeSet(self,symbol):
        self.updateBalance()
        
        iTs = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))
        # redo if uid already reserved
        while iTs in self.tradeSets:
            iTs = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))
        
        ts = {'symbol': symbol,'InTrades': [],'createdAt': time.time(), \
              'OutTrades': [], 'baseCurrency': re.search("(?<=/).*", symbol).group(0), \
              'coinCurrency': re.search(".*(?=/)", symbol).group(0), 'costIn': 0, \
              'costOut': 0, 'coinsAvail': 0, 'initCoins': 0, 'initPrice': None, \
              'SL': None, 'trailingSL': [None, None], 'dailycloseSL': None, \
              'weeklycloseSL': None, 'active': False, 'virgin': True, 'updating': False, \
              'waiting':  []}
        self.tradeSets[iTs] = ts
        return ts, iTs
        
    def activateTradeSet(self,iTs,verbose=True):
        self.updateDownState(True)
        ts = self.tradeSets[iTs]
        wasactive = ts['active']
        # sanity check of amounts to buy/sell
        if self.sumSellAmounts(iTs,'notinitiated') - (self.sumBuyAmounts(iTs,'notfilled')+ ts['coinsAvail']) > 0:
            self.message('Cannot activate trade set because the total amount you (still) want to sell (%s %s) exceeds the total amount you want to buy (%s %s after fee subtraction) and the amount you already have in this trade set (%s %s). Please adjust the trade set!'%(self.amount2Prec(ts['symbol'],self.sumSellAmounts(iTs,'notinitiated',1)),ts['coinCurrency'],self.amount2Prec(ts['symbol'],self.sumBuyAmounts(iTs,'notfilled',1)),ts['coinCurrency'],self.amount2Prec(ts['symbol'],ts['coinsAvail']),ts['coinCurrency']))
            return wasactive
        elif self.minBuyPrice(iTs,order='notfilled') is not None and ts['SL'] is not None and ts['SL'] >= self.minBuyPrice(iTs,order='notfilled'):
            self.message('Cannot activate trade set because the current stop loss price is higher than the lowest non-filled buy order price, which means this buy order could never be reached. Please adjust the trade set!')
            return wasactive
        self.tradeSets[iTs]['virgin'] = False
        self.tradeSets[iTs]['active'] = True
        if verbose and not wasactive:
            totalBuyCost = ts['costIn'] + self.sumBuyCosts(iTs,'notfilled')
            self.message('Estimated return if all trades are executed: %s %s'%(self.cost2Prec(ts['symbol'],self.sumSellCosts(iTs)-totalBuyCost),ts['baseCurrency']))
            if ts['SL'] is not None or ts['dailycloseSL'] is not None:
                sl = [val for val in [ts['SL'] ,ts['dailycloseSL'] if 'dailycloseSL' in ts else None,ts['weeklycloseSL'] if 'weeklycloseSL' in ts else None,] if val is not None][0]
                loss = totalBuyCost - ts['costOut'] - (ts['initCoins']+self.sumBuyAmounts(iTs)-self.sumSellAmounts(iTs,'filled'))*sl
                self.message('Estimated %s if buys reach stop-loss before selling: %s %s'%('*gain*' if loss<0 else 'loss',self.cost2Prec(ts['symbol'],-loss if loss<0 else loss),ts['baseCurrency']))        
        self.initBuyOrders(iTs)
        return wasactive
    
    def deactivateTradeSet(self,iTs,cancelOrders=0):
        self.updateDownState(True)
        # cancelOrders can be 0 (not), 1 (cancel), 2 (cancel and delete open orders)
        wasactive = self.tradeSets[iTs]['active']
        if cancelOrders:
            self.cancelBuyOrders(iTs,deleteOpenOrders=cancelOrders==2)
            self.cancelSellOrders(iTs,deleteOpenOrders=cancelOrders==2)
        self.tradeSets[iTs]['active'] = False
        return wasactive
        
    def newTradeSet(self,symbol,buyLevels=[],buyAmounts=[],sellLevels=[],sellAmounts=[],sl=None,candleAbove=[],initCoins=0,initPrice=None,force=False):
        self.updateDownState(True)
        if symbol not in self.exchange.symbols:
            raise NameError('Trading pair %s not found on %s'%(symbol,self.exchange.name))
        if not self.checkNum(buyAmounts) or not self.checkNum(buyLevels):
            raise TypeError('Buy levels and amounts must be of type float!')
        if not self.checkNum(sellAmounts) or not self.checkNum(sellLevels):
            raise TypeError('Sell levels and amounts must be of type float!')
        if sl and not self.checkNum(sl):
            raise TypeError('Stop-loss must be of type float!')
        if not self.checkNum(initCoins): 
            raise TypeError('Initial coin amount must be of type float!')
        if len(buyLevels)!=len(buyAmounts):
            raise ValueError('The number of buy levels and buy amounts has to be the same')
        if len(sellLevels)!=len(sellAmounts):
            raise ValueError('The number of sell levels and sell amounts has to be the same')
          
        ts, iTs = self.initTradeSet(symbol)

        # truncate values to precision        
        sellLevels = [float(self.exchange.priceToPrecision(ts['symbol'],val)) for val in sellLevels]
        buyLevels = [float(self.exchange.priceToPrecision(ts['symbol'],val)) for val in buyLevels]
        sellAmounts = [float(self.exchange.amountToPrecision(ts['symbol'],val)) for val in sellAmounts]
        buyAmounts = [float(self.exchange.amountToPrecision(ts['symbol'],val)) for val in buyAmounts]

        # sort sell levels and amounts to have lowest level first
        idx = np.argsort(sellLevels)
        sellLevels = np.array(sellLevels)[idx]
        sellAmounts = np.array(sellAmounts)[idx]
        buyLevels = np.array(buyLevels)
        buyAmounts = np.array(buyAmounts)
        if len(buyAmounts) != len(candleAbove):
            candleAbove = np.repeat(None,len(buyAmounts))
        else:
            candleAbove = np.array(candleAbove)

        if not force and sum(buyAmounts) != sum(sellAmounts):
            if self.logger:
                self.logger.warning('It seems the buy and sell amount of %s is not the same. Is this correct?'%ts['coinCurrency'])
        if buyLevels.size > 0 and sellLevels.size > 0 and max(buyLevels) > min(sellLevels):
            raise ValueError('It seems at least one of your sell prices is lower than one of your buy, which does not make sense')
        if self.balance[ts['baseCurrency']]['free'] < sum(buyLevels*buyAmounts):
            raise ValueError('Free balance of %s not sufficient to initiate trade set'%ts['baseCurrency'])
        
        # create the buy orders
        for n,_ in enumerate(buyLevels):
            self.addBuyLevel(iTs,buyLevels[n],buyAmounts[n],candleAbove[n])
        
        self.addInitCoins(iTs,initCoins,initPrice)
        self.setSL(iTs,sl)
        self.setTrailingSL(iTs,None)
        # create the sell orders
        for n,_ in enumerate(sellLevels):
            self.addSellLevel(iTs,sellLevels[n],sellAmounts[n])

        self.activateTradeSet(iTs)
        self.update()
        return iTs
        
    def getTradeSetInfo(self,iTs,showProfitIn=None):
        ts = self.tradeSets[iTs]
        string = '*%srade set #%d on %s%s [%s]:*\n'%('T' if ts['active'] else 'INACTIVE t',list(self.tradeSets.keys()).index(iTs),self.exchange.name,' (DOWN !!!) ' if self.down else '',ts['symbol'])
        filledBuys = []
        filledSells = []
        for iTrade,trade in enumerate(ts['InTrades']):
            tmpstr = '*Buy level %d:* Price %s , Amount %s %s   '%(iTrade,self.price2Prec(ts['symbol'],trade['price']),self.amount2Prec(ts['symbol'],trade['amount']),ts['coinCurrency'])
            if trade['oid'] is None:
                if trade['candleAbove'] is None:
                    tmpstr = tmpstr + '_Order not initiated_\n'
                else:
                    tmpstr = tmpstr + 'if DC > %s'%self.price2Prec(ts['symbol'],trade['candleAbove'])
            elif trade['oid'] == 'filled':
                tmpstr = tmpstr + '_Order filled_\n'
                filledBuys.append([trade['actualAmount'],trade['price']])
            else:
                tmpstr = tmpstr + '_Open order_\n'
            string += tmpstr
        string+= '\n'
        for iTrade,trade in enumerate(ts['OutTrades']):
            tmpstr = '*Sell level %d:* Price %s , Amount %s %s   '%(iTrade,self.price2Prec(ts['symbol'],trade['price']),self.amount2Prec(ts['symbol'],trade['amount']),ts['coinCurrency'])
            if trade['oid'] is None:
                tmpstr = tmpstr + '_Order not initiated_\n'
            elif trade['oid'] == 'filled':
                tmpstr = tmpstr + '_Order filled_\n'
                filledSells.append([trade['amount'],trade['price']])
            else:
                tmpstr = tmpstr + '_Open order_\n'
            string += tmpstr
        if ts['SL'] is not None:
            string += '\n*Stop-loss* set at %s%s\n\n'%(self.price2Prec(ts['symbol'],ts['SL']),'' if ts['trailingSL'][0] is None else (' (trailing with offset %.5g)'%ts['trailingSL'][0] if ts['trailingSL'][1] == 'abs' else ' (trailing with offset %.2g %%)'%(ts['trailingSL'][0]*100) ))
        elif ts['dailycloseSL'] is not None:
            string += '\n*Stop-loss* set at daily close < %s\n\n'%(self.price2Prec(ts['symbol'],ts['dailycloseSL']))
        elif ts['weeklycloseSL'] is not None:
            string += '\n*Stop-loss* set at weekly close < %s\n\n'%(self.price2Prec(ts['symbol'],ts['weeklycloseSL']))
        else:
            string += '\n*No stop-loss set.*\n\n'
        sumBuys = sum([val[0] for val in filledBuys])
        sumSells = sum([val[0] for val in filledSells])
        if ts['initCoins']>0:
            string += '*Initial coins:* %s %s for an average price of %s\n'%(self.amount2Prec(ts['symbol'],ts['initCoins']),ts['coinCurrency'],self.price2Prec(ts['symbol'],ts['initPrice']) if ts['initPrice'] is not None else 'unknown')
        if sumBuys>0:
            string += '*Filled buy orders (fee subtracted):* %s %s for an average price of %s\n'%(self.amount2Prec(ts['symbol'],sumBuys),ts['coinCurrency'],self.cost2Prec(ts['symbol'],sum([val[0]*val[1]/sumBuys if sumBuys > 0 else None for val in filledBuys])))
        if sumSells>0:
            string += '*Filled sell orders:* %s %s for an average price of %s\n'%(self.amount2Prec(ts['symbol'],sumSells),ts['coinCurrency'],self.cost2Prec(ts['symbol'],sum([val[0]*val[1]/sumSells if sumSells > 0 else None for val in filledSells])))
        ticker = self.safeRun(lambda: self.exchange.fetchTicker(ts['symbol']))
        string += '\n*Current market price *: %s, \t24h-high: %s, \t24h-low: %s\n'%tuple([self.price2Prec(ts['symbol'],val) for val in [ticker['last'],ticker['high'],ticker['low']]])
        if (ts['initCoins'] == 0 or ts['initPrice'] is not None) and ts['costIn'] > 0 and (sumBuys>0 or ts['initCoins'] > 0):
            totalAmountToSell = ts['coinsAvail'] + self.sumSellAmounts(iTs,'open')
            fee = self.exchange.calculateFee(ts['symbol'],'market','sell',totalAmountToSell,ticker['last'] ,'taker')
            costSells = ts['costOut'] +   ticker['last'] * totalAmountToSell - (fee['cost'] if fee['currency'] == ts['baseCurrency'] else 0)
            gain = costSells - ts['costIn']
            gainOrig = gain
            if showProfitIn is not None:
                gain,thisCur = self.convertAmount(gain,ts['baseCurrency'],showProfitIn)
            else:
                thisCur = ts['baseCurrency']
            string += '\n*Estimated gain/loss when selling all now: * %s %s (%+.2f %%)\n'%(self.cost2Prec(ts['symbol'],gain),thisCur,gainOrig/(ts['costIn'])*100)
        return string
    

    def createTradeHistoryEntry(self,iTs):
        self.updateDownState(True)
        # create a trade history entry if the trade set had any filled orders
        ts = self.tradeSets[iTs]
        if self.numBuyLevels(iTs,'filled') > 0 or self.numSellLevels(iTs,'filled') > 0:
            gain = ts['costOut']-ts['costIn']
            # try to convert gain amount into btc currency
            gainBTC,curr = self.convertAmount(gain,ts['baseCurrency'],'BTC')
            if curr != 'BTC':
                gainBTC = None
            if gainBTC:
                # try to convert btc gain into usd currency
                gainUSD,curr = self.convertAmount(gainBTC,'BTC','USD')
                if curr != 'USD':
                    gainUSD,curr = self.convertAmount(gainBTC,'BTC','USDT')
                    if curr != 'USDT':
                        gainUSD = None
            else:
                gainUSD = None
            self.tradeSetHistory.append({'time':time.time(),'days':None if not 'createdAt' in ts else (time.time()-ts['createdAt'])/60/60/24 ,'symbol':ts['symbol'],'gain':gain,'gainRel':gain/(ts['costIn'])*100 if ts['costIn']>0 else None,'quote':ts['baseCurrency'],'gainBTC':gainBTC,'gainUSD':gainUSD})
    
    def getTradeHistory(self):
        string = ''
        for tsh in self.tradeSetHistory:
            string += '%s:  %s\t%s%% ( %s BTC | %s USD)\n'%(datetime.datetime.utcfromtimestamp(tsh['time']).strftime('%Y-%m-%d'),tsh['symbol'],'%+7.1f'%tsh['gainRel'] if tsh['gainRel'] else 'N/A','%+.5f'%tsh['gainBTC'] if tsh['gainBTC'] else 'N/A','%+.2f'%tsh['gainUSD'] if tsh['gainUSD'] else 'N/A')
        if len(self.tradeSetHistory) > 0:
            return '*Profit history on %s:\nAvg. relative gain: %+7.1f%%\nTotal profit in BTC: %+.5f\nTotal profit in USD: %+.2f\n\nDetailed Set Info:\n*'%(self.exchange.name,np.mean([tsh['gainRel'] for tsh in self.tradeSetHistory if tsh['gainRel'] is not None]),sum([tsh['gainBTC'] if tsh['gainBTC'] else 0 for tsh in self.tradeSetHistory]),sum([tsh['gainUSD'] if tsh['gainUSD'] else 0 for tsh in self.tradeSetHistory])) + string
        else:
            return '*No profit history on %s*'%self.exchange.name
    
    def resetTradeHistory(self):
        self.tradeSetHistory = []
        return 1
    
    def convertAmount(self,amount,currency,targetCurrency):
        self.updateDownState(True)
        if isinstance(targetCurrency,str):
            targetCurrency = [targetCurrency]
        conversionPairs = [('%s/%s'%(currency,cur) in self.exchange.symbols) + 2*('%s/%s'%(cur,currency) in self.exchange.symbols) for cur in targetCurrency]
        ind = next((i for i, x in enumerate(conversionPairs) if x), None)
        if ind is not None:
            thisCur = targetCurrency[ind]
            if conversionPairs[ind] == 1:
                amount *= self.safeRun(lambda: self.exchange.fetchTicker('%s/%s'%(currency,thisCur)))['last']
            else:
                amount /= self.safeRun(lambda: self.exchange.fetchTicker('%s/%s'%(thisCur,currency)))['last']
            return (amount,thisCur)
        else:
            return (amount,currency)
        
        
    def deleteTradeSet(self,iTs,sellAll=False):
        self.updateDownState(True)
        self.lockTradeSet(iTs)
        if sellAll:
            sold = self.sellAllNow(iTs)
        else:
            sold = True
            self.deactivateTradeSet(iTs,1)
        if sold:
            self.createTradeHistoryEntry(iTs)
            self.tradeSets.pop(iTs)
        else:
            self.unlockTradeSet(iTs)
    
    def addInitCoins(self,iTs,initCoins=0,initPrice=None):
        if self.checkNum(initCoins,initPrice) or (initPrice is None and self.checkNum(initCoins)):
            if initPrice is not None and initPrice < 0:
                initPrice = None
            ts = self.tradeSets[iTs]
            # check if free balance is indeed sufficient
            if self.getFreeBalance(ts['coinCurrency']) < initCoins:
                self.message('Adding initial balance failed: %s %s requested but only %s %s are free!'%(self.amount2Prec(ts['symbol'],initCoins),ts['coinCurrency'],self.amount2Prec(ts['symbol'],self.getFreeBalance(ts['coinCurrency'])),ts['coinCurrency']),'error')
                return 0
            self.lockTradeSet(iTs)
            
            if ts['coinsAvail'] > 0 and ts['initPrice'] is not None:
                # remove old cost again
                ts['costIn'] -= (ts['coinsAvail']*ts['initPrice'])
            ts['coinsAvail'] = initCoins
            ts['initCoins'] = initCoins
            
            ts['initPrice'] = initPrice
            if initPrice is not None:
                ts['costIn'] += (initCoins*initPrice)
            self.unlockTradeSet(iTs)
            return 1
        else:
            raise ValueError('Some input was no number')
            
    def numBuyLevels(self,iTs, order='all'):
        return self.getTradeParam(iTs,'amount','num','buy',order)
    
    def numSellLevels(self,iTs, order='all'):
        return self.getTradeParam(iTs,'amount','num','sell',order)

    def sumBuyAmounts(self,iTs,order='all',subtractFee = True):
        return self.getTradeParam(iTs,'amount','sum','buy',order,subtractFee)
    
    def sumSellAmounts(self,iTs,order='all',subtractFee = True):
        return self.getTradeParam(iTs,'amount','sum','sell',order,subtractFee)
    
    def sumBuyCosts(self,iTs,order='all',subtractFee = True):
        return self.getTradeParam(iTs,'cost','sum','buy',order,subtractFee)
    
    def sumSellCosts(self,iTs,order='all',subtractFee = True):
        return self.getTradeParam(iTs,'cost','sum','sell',order,subtractFee)
    
    def minBuyPrice(self,iTs,order='all'):
        return self.getTradeParam(iTs,'price','min','buy',order)
    
    def getTradeParam(self,iTs,what,method,direction, order='all',subtractFee = True):
        if method == 'sum':
            func = lambda x: sum(x)
        elif method == 'min':
            func = lambda x: None if len(x)==0 else np.min(x)
        elif method == 'max':
            func = lambda x: None if len(x)==0 else np.max(x)
        elif method == 'mean':
            func = lambda x: None if len(x)==0 else np.mean(x)
        elif method == 'num':
            func = lambda x: len(x)
            
        if direction == 'sell':
            trade = 'OutTrades'
        else:
            trade = 'InTrades'
            
        if order not in ['all','filled','open','notfilled','notinitiated']:
            raise ValueError('order has to be all, filled, notfilled, notinitiated or open')
            
        if what == 'amount':
            if order == 'all':
                return func([(val['amount'] if direction == 'sell' or subtractFee == False else val['actualAmount']) for val in self.tradeSets[iTs][trade]])
            elif order == 'filled':
                return func([(val['amount'] if direction == 'sell' or subtractFee == False else val['actualAmount'])  for val in self.tradeSets[iTs][trade] if val['oid'] == 'filled'])
            elif order == 'open':
                return func([(val['amount'] if direction == 'sell' or subtractFee == False else val['actualAmount'])  for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled' and val['oid'] is not None])
            elif order == 'notinitiated':
                return func([(val['amount'] if direction == 'sell' or subtractFee == False else val['actualAmount'])  for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled' and val['oid'] is None])
            elif order == 'notfilled':
                return func([(val['amount'] if direction == 'sell' or subtractFee == False else val['actualAmount'])  for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled'])
        elif what == 'price':
            if order == 'all':
                return func([val['price'] for val in self.tradeSets[iTs][trade]])
            elif order == 'filled':
                return func([val['price'] for val in self.tradeSets[iTs][trade] if val['oid'] == 'filled'])
            elif order == 'open':
                return func([val['price'] for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled' and val['oid'] is not None])
            elif order == 'notinitiated':
                return func([val['price'] for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled' and val['oid'] is None])
            elif order == 'notfilled':
                return func([val['price'] for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled'])   
        elif what == 'cost':
            if order == 'all':
                return func([val['amount']*val['price'] for val in self.tradeSets[iTs][trade]])
            elif order == 'filled':
                return func([val['amount']*val['price'] for val in self.tradeSets[iTs][trade] if val['oid'] == 'filled'])
            elif order == 'open':
                return func([val['amount']*val['price'] for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled' and val['oid'] is not None])
            elif order == 'notinitiated':
                return func([val['amount']*val['price'] for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled' and val['oid'] is None])
            elif order == 'notfilled':
                return func([val['amount']*val['price'] for val in self.tradeSets[iTs][trade] if val['oid'] != 'filled'])   
            
    
    def addBuyLevel(self,iTs,buyPrice,buyAmount,candleAbove=None):
        self.updateDownState(True)
        ts = self.tradeSets[iTs]
        if self.checkNum(buyPrice,buyAmount,candleAbove) or (candleAbove is None and self.checkNum(buyPrice,buyAmount)):
            fee = self.exchange.calculateFee(ts['symbol'],'limit','buy',buyAmount,buyPrice,'maker')
            if not self.checkQuantity(ts['symbol'],'amount',buyAmount):
                self.message('Adding buy level failed, amount is not within the range, the exchange accepts')
                return 0
            elif not self.checkQuantity(ts['symbol'],'price',buyPrice):
                self.message('Adding buy level failed, price is not within the range, the exchange accepts')
                return 0 
            elif not self.checkQuantity(ts['symbol'],'cost',buyPrice*buyAmount):
                self.message('Adding buy level failed, cost is not within the range, the exchange accepts')
                return 0 
            elif self.getFreeBalance(ts['baseCurrency']) < buyAmount*buyPrice + fee['cost'] if fee['currency'] == ts['baseCurrency'] else 0:
                self.message('Adding buy level failed, your balance of %s does not suffice to buy this amount%s!'%(ts['baseCurrency'],' and pay the trading fee (%s %s)'%(self.fee2Prec(ts['symbol'],fee['cost']),ts['baseCurrency']) if fee['currency'] == ts['baseCurrency'] else ''))
                return 0 
            
            if fee['currency'] == ts['coinCurrency']:
                boughtAmount = buyAmount - (fee['cost'] if (self.exchange.name.lower() != 'binance' or self.getFreeBalance('BNB') < 0.5) else 0) # this is a hack, as fees on binance are deduced from BNB if this is activated and there is enough BNB, however so far no API chance to see if this is the case. Here I assume that 0.5 BNB are enough to pay the fee for the trade and thus the fee is not subtracted from the traded coin
            else:
                boughtAmount = buyAmount
            self.lockTradeSet(iTs)
            wasactive = self.deactivateTradeSet(iTs)  
            ts['InTrades'].append({'oid': None, 'price': buyPrice, 'amount': buyAmount, 'actualAmount': boughtAmount, 'candleAbove': candleAbove})
            if wasactive:
                self.activateTradeSet(iTs,0)   
            self.unlockTradeSet(iTs)
            return  self.numBuyLevels(iTs)-1
        else:
            raise ValueError('Some input was no number')
    
    def deleteBuyLevel(self,iTs,iTrade):
        self.updateDownState(True)
        if self.checkNum(iTrade):
            self.lockTradeSet(iTs)
            ts = self.tradeSets[iTs]
            wasactive = self.deactivateTradeSet(iTs)
            if ts['InTrades'][iTrade]['oid'] is not None and ts['InTrades'][iTrade]['oid'] != 'filled' :
                self.cancelBuyOrders(iTs,ts['InTrades'][iTrade]['oid'])
            ts['InTrades'].pop(iTrade)
            if wasactive:
                self.activateTradeSet(iTs,0) 
            self.unlockTradeSet(iTs)
        else:
            raise ValueError('Some input was no number')
            
    def setBuyLevel(self,iTs,iTrade,price,amount): 
        self.updateDownState(True)
        if self.checkNum(iTrade,price,amount):
            ts = self.tradeSets[iTs]
            if ts['InTrades'][iTrade]['oid'] == 'filled':
                self.message('This order is already filled! No change possible')
                return 0
            else:
                fee = self.exchange.calculateFee(ts['symbol'],'limit','buy',amount,price,'maker')
                if not self.checkQuantity(ts['symbol'],'amount',amount):
                    self.message('Changing buy level failed, amount is not within the range, the exchange accepts')
                    return 0
                elif not self.checkQuantity(ts['symbol'],'price',price):
                    self.message('Changing buy level failed, price is not within the range, the exchange accepts')
                    return 0 
                elif not self.checkQuantity(ts['symbol'],'cost',price*amount):
                    self.message('Changing buy level failed, cost is not within the range, the exchange accepts')
                    return 0 
                elif self.getFreeBalance(ts['baseCurrency']) + ts['InTrades'][iTrade]['amount']*ts['InTrades'][iTrade]['price'] < amount*price + fee['cost'] if fee['currency'] == ts['baseCurrency'] else 0:
                    self.message('Changing buy level failed, your balance of %s does not suffice to buy this amount%s!'%(ts['baseCurrency'],' and pay the trading fee (%s %s)'%(self.fee2Prec(ts['symbol'],fee['cost']),ts['baseCurrency']) if fee['currency'] == ts['baseCurrency'] else ''))
                    return 0 
                if fee['currency'] == ts['coinCurrency']:
                    boughtAmount = amount - (fee['cost'] if (self.exchange.name.lower() != 'binance' or self.getFreeBalance('BNB') < 0.5) else 0) # this is a hack, as fees on binance are deduced from BNB if this is activated and there is enough BNB, however so far no API chance to see if this is the case. Here I assume that 0.5 BNB are enough to pay the fee for the trade and thus the fee is not subtracted from the traded coin
                else:
                    boughtAmount = amount
                wasactive = self.deactivateTradeSet(iTs)  
                
                if ts['InTrades'][iTrade]['oid'] is not None and ts['InTrades'][iTrade]['oid'] != 'filled' :
                    returnVal = self.cancelBuyOrders(iTs,ts['InTrades'][iTrade]['oid'])
                    ts['InTrades'][iTrade]['oid'] = None
                if returnVal == 0.5 and self.getFreeBalance(ts['baseCurrency']) + ts['InTrades'][iTrade]['amount']*ts['InTrades'][iTrade]['price'] < amount*price + fee['cost'] if fee['currency'] == ts['baseCurrency'] else 0:
                    self.message('Changing buy level failed, your balance of %s does not suffice to buy this amount%s!'%(ts['baseCurrency'],' and pay the trading fee (%s %s)'%(self.fee2Prec(ts['symbol'],fee['cost']),ts['baseCurrency']) if fee['currency'] == ts['baseCurrency'] else ''))
                    return 0
                ts['InTrades'][iTrade].update({'amount': amount, 'actualAmount': boughtAmount, 'price': price})
                    
                if wasactive:
                    self.activateTradeSet(iTs,0)                
                return 1
        else:
            raise ValueError('Some input was no number')
    
    def addSellLevel(self,iTs,sellPrice,sellAmount):
        self.updateDownState(True)
        ts = self.tradeSets[iTs]
        if self.checkNum(sellPrice,sellAmount):
            if not self.checkQuantity(ts['symbol'],'amount',sellAmount):
                self.message('Adding sell level failed, amount is not within the range, the exchange accepts')
                return 0
            elif not self.checkQuantity(ts['symbol'],'price',sellPrice):
                self.message('Adding sell level failed, price is not within the range, the exchange accepts')
                return 0 
            elif not self.checkQuantity(ts['symbol'],'cost',sellPrice*sellAmount):
                self.message('Adding sell level failed, return is not within the range, the exchange accepts')
                return 0 
            self.lockTradeSet(iTs)
            wasactive = self.deactivateTradeSet(iTs)  
            ts['OutTrades'].append({'oid': None, 'price': sellPrice, 'amount': sellAmount})
            if wasactive:
                self.activateTradeSet(iTs,0)  
            self.unlockTradeSet(iTs)
            return  self.numSellLevels(iTs)-1
        else:
            raise ValueError('Some input was no number')

    def deleteSellLevel(self,iTs,iTrade):
        self.updateDownState(True)
        if self.checkNum(iTrade):
            self.lockTradeSet(iTs)
            ts = self.tradeSets[iTs]
            wasactive = self.deactivateTradeSet(iTs)
            if ts['OutTrades'][iTrade]['oid'] is not None and ts['OutTrades'][iTrade]['oid'] != 'filled' :
                self.cancelSellOrders(iTs,ts['OutTrades'][iTrade]['oid'])
            ts['OutTrades'].pop(iTrade)
            self.unlockTradeSet(iTs)
            if wasactive:
                self.activateTradeSet(iTs,0) 
        else:
            raise ValueError('Some input was no number')
    
    def setSellLevel(self,iTs,iTrade,price,amount):  
        self.updateDownState(True)
        if self.checkNum(iTrade,price,amount):
            ts = self.tradeSets[iTs]
            if ts['OutTrades'][iTrade]['oid'] == 'filled':
                self.message('This order is already filled! No change possible')
                return 0
            else:
                if not self.checkQuantity(ts['symbol'],'amount',amount):
                    self.message('Changing sell level failed, amount is not within the range, the exchange accepts')
                    return 0
                elif not self.checkQuantity(ts['symbol'],'price',price):
                    self.message('Changing sell level failed, price is not within the range, the exchange accepts')
                    return 0 
                elif not self.checkQuantity(ts['symbol'],'cost',price*amount):
                    self.message('Changing sell level failed, return is not within the range, the exchange accepts')
                    return 0 
                wasactive = self.deactivateTradeSet(iTs)  
                
                if ts['OutTrades'][iTrade]['oid'] is not None and ts['OutTrades'][iTrade]['oid'] != 'filled' :
                    self.cancelSellOrders(iTs,ts['OutTrades'][iTrade]['oid'])
                    ts['OutTrades'][iTrade]['oid'] = None

                ts['OutTrades'][iTrade]['amount'] = amount
                ts['OutTrades'][iTrade]['price'] = price
                
                if wasactive:
                    self.activateTradeSet(iTs,0)                
                return 1
        else:
            raise ValueError('Some input was no number')
            
    def setTrailingSL(self,iTs,value,typ='abs'): 
        self.updateDownState(True)
        ts = self.tradeSets[iTs]
        if self.checkNum(value):
            if self.numBuyLevels(iTs,'notfilled') > 0:
                raise Exception('Trailing SL cannot be set as there are non-filled buy orders still')
            ticker = self.safeRun(lambda: self.exchange.fetch_ticker(ts['symbol']))
            if typ == 'abs':
                if value >= ticker['last'] or value <= 0:
                    raise ValueError('absolute trailing stop-loss offset is not between 0 and current price')
                newSL = ticker['last'] - value
            else:
                if value >= 1 or value <= 0:
                    raise ValueError('Relative trailing stop-loss offset is not between 0 and 1')
                newSL = ticker['last'] * (1- value)     
            ts['trailingSL'] = [value,typ]          
            ts['SL'] = newSL
            self.setDailyCloseSL(iTs,None)
            self.setWeeklyCloseSL(iTs,None)
        elif value is None:
            ts['trailingSL'] = [None,None]
        else:
            raise ValueError('Input was no number')
            
    def setWeeklyCloseSL(self,iTs,value):
        ts = self.tradeSets[iTs]
        if self.checkNum(value):
            ts['weeklycloseSL'] = value
            self.setDailyCloseSL(iTs,None)
            self.setTrailingSL(iTs,None) # deactivate trailing SL
            self.setSL(iTs,None)  # deactivate standard SL
        elif value is None:
            ts['weeklycloseSL'] = None
        else:
            raise ValueError('Input was no number')    
            
    def setDailyCloseSL(self,iTs,value):
        ts = self.tradeSets[iTs]
        if self.checkNum(value):
            ts['dailycloseSL'] = value
            self.setWeeklyCloseSL(iTs,None)
            self.setTrailingSL(iTs,None) # deactivate trailing SL
            self.setSL(iTs,None)  # deactivate standard SL
        elif value is None:
            ts['dailycloseSL'] = None
        else:
            raise ValueError('Input was no number')    
    
    def setSL(self,iTs,value):   
        if self.checkNum(value):
            ts = self.tradeSets[iTs]
            ticker = self.safeRun(lambda: self.exchange.fetch_ticker(ts['symbol']))
            if ticker['last'] <= value:
                self.message('Cannot set new SL as it is higher than the current market price')
                return 0
            self.setTrailingSL(iTs,None) # deactivate trailing SL
            self.setDailyCloseSL(iTs,None) # same for dailyclose SL
            self.setWeeklyCloseSL(iTs,None) # same for weeklycloseSL SL
            self.tradeSets[iTs]['SL'] = value
            return 1
        elif value is None:
            self.tradeSets[iTs]['SL'] = None
            return 1
        else:
            raise ValueError('Input was no number')
        
    def setSLBreakEven(self,iTs): 
        self.updateDownState(True)
        ts = self.tradeSets[iTs]         
        if ts['initCoins'] > 0 and ts['initPrice'] is None:
            self.message('Break even SL cannot be set as you this trade set contains %s that you obtained beforehand and no buy price information was given.'%ts['coinCurrency'])
            return 0
        elif ts['costOut'] - ts['costIn'] > 0:
            self.message('Break even SL cannot be set as your sold coins of this trade already outweigh your buy expenses (congrats!)! You might choose to sell everything immediately if this is what you want.')
            return 0
        elif ts['costOut'] - ts['costIn']  == 0:
            self.message('Break even SL cannot be set as there are no unsold %s coins right now'%ts['coinCurrency'])
            return 0
        else:
            breakEvenPrice = (ts['costIn']-ts['costOut'])/((1-self.exchange.fees['trading']['taker'])*(ts['coinsAvail']+sum([trade['amount'] for trade in ts['OutTrades'] if trade['oid'] != 'filled' and trade['oid'] is not None])))
            ticker = self.safeRun(lambda :self.exchange.fetch_ticker(ts['symbol']))
            if ticker['last'] < breakEvenPrice:
                self.message('Break even SL of %s cannot be set as the current market price is lower (%s)!'%tuple([self.price2Prec(ts['symbol'],val) for val in [breakEvenPrice,ticker['last']]]))
                return 0
            else:
                self.setSL(iTs,breakEvenPrice)
                return 1

    def sellAllNow(self,iTs,price=None):
        self.updateDownState(True)
        self.deactivateTradeSet(iTs,2)
        ts = self.tradeSets[iTs]
        ts['SL'] = None # necessary to not retrigger SL
        ts['dailycloseSL'] = None # this one, too
        sold = True
                
        if ts['coinsAvail'] > 0 and self.checkQuantity(ts['symbol'],'amount',ts['coinsAvail']):
            if self.exchange.has['createMarketOrder']:
                try:
                    response = self.safeRun(lambda: self.exchange.createMarketSellOrder (ts['symbol'], ts['coinsAvail']),0)
                except:
                    params = { 'trading_agreement': 'agree' }  # for kraken api...
                    response = self.safeRun(lambda: self.exchange.createMarketSellOrder (ts['symbol'], ts['coinsAvail'],params),iTs=iTs)
            else:
                if price is None:
                    price = self.safeRun(lambda :self.exchange.fetch_ticker(ts['symbol'])['last'], iTs=iTs)
                response = self.safeRun(lambda: self.exchange.createLimitSellOrder (ts['symbol'], ts['coinsAvail'],price),iTs=iTs)
            time.sleep(5) # give exchange 5 sec for trading the order
            orderInfo = self.fetchOrder(response['id'],iTs,'SELL')
                    
            if orderInfo['status']=='FILLED':
                if orderInfo['type'] == 'market' and self.exchange.has['fetchMyTrades'] != False:
                    trades = self.exchange.fetchMyTrades(ts['symbol'])
                    orderInfo['cost'] = sum([tr['cost'] for tr in trades if tr['order'] == orderInfo['id']])
                    orderInfo['price'] = np.mean([tr['price'] for tr in trades if tr['order'] == orderInfo['id']])
                ts['costOut'] += orderInfo['cost']
                self.message('Sold immediately at a price of %s %s: Sold %s %s for %s %s.'%(self.price2Prec(ts['symbol'],orderInfo['price']),ts['symbol'],self.amount2Prec(ts['symbol'],orderInfo['amount']),ts['coinCurrency'],self.cost2Prec(ts['symbol'],orderInfo['cost']),ts['baseCurrency']))
            else:
                self.message('Sell order was not traded immediately, updating status soon.')
                sold = False
                ts['OutTrades'].append({'oid':response['id'],'price': orderInfo['price'],'amount': orderInfo['amount']})
                self.activateTradeSet(iTs,0)
        else:
            self.message('No coins (or too low amount) to sell from this trade set.','warning')
        return sold
                
    def cancelSellOrders(self,iTs,oid=None,deleteOpenOrders=False):
        self.updateDownState(True)
        returnVal = 1
        if iTs in self.tradeSets and self.numSellLevels(iTs) > 0:
            count = 0
            for iTrade,trade in reversed(list(enumerate(self.tradeSets[iTs]['OutTrades']))):
                if trade['oid'] is not None and trade['oid'] != 'filled' and (oid is None or trade['oid']==oid):
                    try:
                        self.cancelOrder(trade['oid'],iTs,'SELL') 
                    except OrderNotFound as e:
                        pass
                    except Exception as e:
                        self.unlockTradeSet(iTs)
                        raise(e)
                    time.sleep(1)
                    count += 1
                    orderInfo = self.fetchOrder(trade['oid'],iTs,'SELL')
                    self.tradeSets[iTs]['coinsAvail'] += trade['amount']
                    trade['oid'] = None
                    if orderInfo['filled'] > 0:
                        self.message('(Partly?) filled sell order found during canceling. Updating balance')
                        self.tradeSets[iTs]['costOut'] += orderInfo['price']*orderInfo['filled']
                        self.tradeSets[iTs]['coinsAvail'] -= orderInfo['filled']  
                        trade['oid'] = 'filled'
                        trade['amount'] = orderInfo['filled']
                        returnVal = 0.5
                    elif deleteOpenOrders:
                        self.tradeSets[iTs]['OutTrades'].pop(iTrade)
            if count > 0:
                self.message('%d sell orders canceled in total for tradeSet %d (%s)'%(count,list(self.tradeSets.keys()).index(iTs),self.tradeSets[iTs]['symbol']))
        return returnVal
        
    def cancelBuyOrders(self,iTs,oid=None,deleteOpenOrders=False):
        self.updateDownState(True)
        returnVal = 1
        if iTs in self.tradeSets and self.numBuyLevels(iTs) > 0:
            count = 0
            for iTrade,trade in reversed(list(enumerate(self.tradeSets[iTs]['InTrades']))):
                if trade['oid'] is not None and trade['oid'] != 'filled' and (oid is None or trade['oid']==oid):
                    try:
                        self.cancelOrder(trade['oid'],iTs,'BUY') 
                    except OrderNotFound as e:
                        pass
                        
                    time.sleep(1)
                    count += 1
                    orderInfo = self.fetchOrder(trade['oid'],iTs,'BUY')
                    trade['oid'] = None
                    if orderInfo['filled'] > 0:
                        self.message('(Partly?) filled buy order found during canceling. Updating balance')
                        self.tradeSets[iTs]['costIn'] += orderInfo['price']*orderInfo['filled']
                        self.tradeSets[iTs]['coinsAvail'] += orderInfo['filled']   
                        returnVal = 0.5
                    elif deleteOpenOrders:
                        self.tradeSets[iTs]['InTrades'].pop(iTrade)
            if count > 0:
                self.message('%d buy orders canceled in total for tradeSet %d (%s)'%(count,list(self.tradeSets.keys()).index(iTs),self.tradeSets[iTs]['symbol']))
        return returnVal
    
    def initBuyOrders(self,iTs):
        self.updateDownState(True)
        if self.tradeSets[iTs]['active']:
            # initialize buy orders
            for iTrade,trade in enumerate(self.tradeSets[iTs]['InTrades']):
                if trade['oid'] is None and trade['candleAbove'] is None:
                    response = self.safeRun(lambda: self.exchange.createLimitBuyOrder(self.tradeSets[iTs]['symbol'], trade['amount'],trade['price']))
                    self.tradeSets[iTs]['InTrades'][iTrade]['oid'] = response['id']
    
    def cancelOrder(self,oid,iTs,typ):
        self.updateDownState(True)
        symbol = self.tradeSets[iTs]['symbol']
        try:
            return self.safeRun(lambda: self.exchange.cancelOrder (oid,symbol),0)
        except OrderNotFound as e:
            self.unlockTradeSet(iTs)
            raise(e)
        except ccxt.ExchangeError as e:
            return self.safeRun(lambda: self.exchange.cancelOrder (oid,symbol,{'type':typ}) , iTs=iTs)
        
    def fetchOrder(self,oid,iTs,typ):
        symbol = self.tradeSets[iTs]['symbol']
        try:
            return self.safeRun(lambda: self.exchange.fetchOrder (oid,symbol),0)
        except OrderNotFound as e:
            self.unlockTradeSet(iTs)
            raise(e)
        except ccxt.ExchangeError as e:
            return self.safeRun(lambda: self.exchange.fetchOrder (oid,symbol,{'type':typ}), iTs=iTs)  
        
    def updateDownState(self,raiseError = False):
        if self.down:
            self.safeRun(self.exchange.loadMarkets, printError = False)
            if raiseError:
                raise ccxt.ExchangeError('Exchange is down!')
            return self.down
        else:
            return False
        
    def update(self,specialCheck=0):
        # goes through all trade sets and checks/updates the buy/sell/stop loss orders
        # daily check is for checking if a candle closed above a certain value
        if self.updateDownState():
            # check if exchange is still down, if yes, return
            return
        else: 
            try:
                self.updateBalance()
            except AuthenticationError as e:#
                self.message('Failed to authenticate at exchange %s. Please check your keys'%self.exchange.name,'error')
                return
            except ccxt.ExchangeError as e:#
                if 'key' in str(e).lower():
                    self.message('Failed to authenticate at exchange %s. Please check your keys'%self.exchange.name,'error')
                else:
                    self.message('Some error occured at exchange %s. Maybe it is down.'%self.exchange.name,'error')
                return
        
        tradeSetsToDelete = []
        try:
            for iTs in self.tradeSets:
                ts = self.tradeSets[iTs]
                if not ts['active']:
                    continue
                self.lockTradeSet(iTs)
                ticker = self.safeRun(lambda: self.exchange.fetch_ticker(ts['symbol']), iTs=iTs)
                # check if stop loss is reached
                if not specialCheck:
                    if ts['SL'] is not None:
                        if ticker['last'] <= ts['SL']:
                            self.message('Stop loss for pair %s has been triggered!'%ts['symbol'],'warning')
                            # cancel all sell orders, create market sell order and save resulting amount of base currency
                            sold = self.sellAllNow(iTs,price=ticker['last'])
                            if sold:
                                tradeSetsToDelete.append(iTs)
                                self.unlockTradeSet(iTs)
                                continue
                        elif 'trailingSL' in ts and ts['trailingSL'][0] is not None:
                            if ts['trailingSL'][1] == 'abs':
                                newSL = ticker['last'] - ts['trailingSL'][0]
                            else:
                                newSL = ticker['last'] * (1- ts['trailingSL'][0])
                            if newSL > ts['SL']:
                                ts['SL'] = newSL
                elif specialCheck == 1 and 'dailycloseSL' in ts and ts['dailycloseSL'] is not None and ticker['last'] < ts['dailycloseSL']:
                    self.message('Daily candle closed below chosen SL of %s for pair %s! Selling now!'%(self.price2Prec(ts['symbol'],ts['dailycloseSL']),ts['symbol']),'warning')
                    # cancel all sell orders, create market sell order and save resulting amount of base currency
                    sold = self.sellAllNow(iTs,price=ticker['last'])
                    if sold:
                        tradeSetsToDelete.append(iTs)
                        self.unlockTradeSet(iTs)
                        continue                                    
                elif specialCheck == 2 and 'weeklycloseSL' in ts and ts['weeklycloseSL'] is not None and ticker['last'] < ts['weeklycloseSL']:
                    self.message('Weekly candle closed below chosen SL of %s for pair %s! Selling now!'%(self.price2Prec(ts['symbol'],ts['weeklycloseSL']),ts['symbol']),'warning')
                    # cancel all sell orders, create market sell order and save resulting amount of base currency
                    sold = self.sellAllNow(iTs,price=ticker['last'])
                    if sold:
                        tradeSetsToDelete.append(iTs)
                        self.unlockTradeSet(iTs)
                        continue
                orderExecuted = 0
                # go through buy trades 
                for iTrade,trade in enumerate(ts['InTrades']):
                    if trade['oid'] == 'filled':
                        continue
                    elif specialCheck == 1 and trade['oid'] is None and trade['candleAbove'] is not None:
                        if ticker['last'] > trade['candleAbove']:
                            response = self.safeRun(lambda: self.exchange.createLimitBuyOrder(ts['symbol'], trade['amount'],trade['price']), iTs=iTs)
                            ts['InTrades'][iTrade]['oid'] = response['id']
                            self.message('Daily candle of %s above %s triggering buy level #%d on %s!'%(ts['symbol'],self.price2Prec(ts['symbol'],trade['candleAbove']),iTrade,self.exchange.name))
                    elif trade['oid'] is not None:
                        orderInfo = self.fetchOrder(trade['oid'],iTs,'BUY')
                        # fetch trades for all orders because a limit order might also be filled at a lower val
                        if self.exchange.has['fetchMyTrades'] != False:
                            trades = self.exchange.fetchMyTrades(ts['symbol'])
                            orderInfo['cost'] = sum([tr['cost'] for tr in trades if tr['order'] == orderInfo['id']])
                            if orderInfo['cost'] == 0:
                                orderInfo['price'] = None
                            else:
                                orderInfo['price'] = np.mean([tr['price'] for tr in trades if tr['order'] == orderInfo['id']])
                        if any([orderInfo['status'].lower() == val for val in ['closed','filled']]):
                            orderExecuted = 1
                            ts['InTrades'][iTrade]['oid'] = 'filled'
                            
                            ts['InTrades'][iTrade]['price'] = orderInfo['price']
                            ts['costIn'] += orderInfo['cost']
                            self.message('Buy level of %s %s reached on %s! Bought %s %s for %s %s.'%(self.price2Prec(ts['symbol'],orderInfo['price']),ts['symbol'],self.exchange.name,self.amount2Prec(ts['symbol'],orderInfo['amount']),ts['coinCurrency'],self.cost2Prec(ts['symbol'],orderInfo['cost']),ts['baseCurrency']))
                            ts['coinsAvail'] += trade['actualAmount']
                        elif orderInfo['status'] == 'canceled':
                            if 'reason' in orderInfo['info']:
                                reason = orderInfo['info']['reason']
                            else:
                                reason = 'N/A'
                            if orderInfo['cost'] > 0:
                                ts['InTrades'][iTrade]['oid'] = 'filled'
                                ts['InTrades'][iTrade]['price'] = orderInfo['price']
                                ts['costIn'] += orderInfo['cost']  
                                orderInfo['amount'] = sum([tr['amount'] for tr in trades if tr['order'] == orderInfo['id']])
                                ts['coinsAvail'] += orderInfo['amount']
                                self.message('Buy order (level %d of trade set %d on %s) was canceled by exchange or someone else (reason: %s) but already partly filled! Treating order as closed now and updating trade set info.'%(iTrade,list(self.tradeSets.keys()).index(iTs),self.exchange.name))
                            else:
                                ts['InTrades'][iTrade]['oid'] = None
                                self.message('Buy order (level %d of trade set %d on %s) was canceled by exchange or someone else (reason:%s)! Will be reinitialized during next update.'%(iTrade,list(self.tradeSets.keys()).index(iTs),self.exchange.name,reason))
                    else:
                        self.initBuyOrders(iTs)                                
                        time.sleep(1)
                if not specialCheck:
                    # go through all selling positions and create those for which the bought coins suffice
                    for iTrade,_ in enumerate(ts['OutTrades']):
                        if ts['OutTrades'][iTrade]['oid'] is None and ts['coinsAvail'] >= ts['OutTrades'][iTrade]['amount']:
                            response = self.safeRun(lambda: self.exchange.createLimitSellOrder(ts['symbol'], ts['OutTrades'][iTrade]['amount'], ts['OutTrades'][iTrade]['price']), iTs=iTs)
                            ts['OutTrades'][iTrade]['oid'] = response['id']
                            ts['coinsAvail'] -= ts['OutTrades'][iTrade]['amount']
                    # go through sell trades 
                    for iTrade,trade in enumerate(ts['OutTrades']):
                        if trade['oid'] == 'filled':
                            continue
                        elif trade['oid'] is not None:
                            orderInfo = self.fetchOrder(trade['oid'],iTs,'SELL')
                            # fetch trades for all orders because a limit order might also be filled at a higher val
                            if self.exchange.has['fetchMyTrades'] != False:
                                trades = self.exchange.fetchMyTrades(ts['symbol'])
                                orderInfo['cost'] = sum([tr['cost'] for tr in trades if tr['order'] == orderInfo['id']])
                                if orderInfo['cost'] == 0:
                                    orderInfo['price'] = None
                                else:
                                    orderInfo['price'] = np.mean([tr['price'] for tr in trades if tr['order'] == orderInfo['id']])
                            if any([orderInfo['status'].lower() == val for val in ['closed','filled']]):
                                orderExecuted = 2
                                ts['OutTrades'][iTrade]['oid'] = 'filled'                               
                                ts['OutTrades'][iTrade]['price'] = orderInfo['price']
                                ts['costOut'] += orderInfo['cost']
                                self.message('Sell level of %s %s reached on %s! Sold %s %s for %s %s.'%(self.price2Prec(ts['symbol'],orderInfo['price']),ts['symbol'],self.exchange.name,self.amount2Prec(ts['symbol'],orderInfo['amount']),ts['coinCurrency'],self.cost2Prec(ts['symbol'],orderInfo['cost']),ts['baseCurrency']))
                            elif orderInfo['status'] == 'canceled':
                                if 'reason' in orderInfo['info']:
                                    reason = orderInfo['info']['reason']
                                else:
                                    reason = 'N/A'
                                if orderInfo['cost'] > 0:
                                    ts['OutTrades'][iTrade]['oid'] = 'filled'
                                    ts['OutTrades'][iTrade]['price'] = orderInfo['price']
                                    ts['costOut'] += orderInfo['cost']    
                                    orderInfo['amount'] = sum([tr['amount'] for tr in trades if tr['order'] == orderInfo['id']])
                                    ts['coinsAvail'] +=  ts['OutTrades'][iTrade]['amount']-orderInfo['amount']
                                    self.message('Sell order (level %d of trade set %d on %s) was canceled by exchange or someone else (reason: %s) but already partly filled! Treating order as closed now and updating trade set info.'%(iTrade,list(self.tradeSets.keys()).index(iTs),self.exchange.name))
                                else:
                                    ts['OutTrades'][iTrade]['oid'] = None
                                    self.message('Sell order (level %d of trade set %d on %s) was canceled by exchange or someone else (reason:%s)! Will be reinitialized during next update.'%(iTrade,list(self.tradeSets.keys()).index(iTs),self.exchange.name,reason))
                    # delete Tradeset when all orders have been filled (but only if there were any to execute and if no significant coin amount is left)
                    if ((ts['SL'] is None and orderExecuted > 0) or (orderExecuted == 2 and (self.sumBuyAmounts(iTs,order='filled')+ts['initCoins']-self.sumSellAmounts(iTs,order='filled'))/(ts['initCoins']+self.sumBuyAmounts(iTs,order='filled')) < 0.01 )) and self.numSellLevels(iTs,'notfilled') == 0 and self.numBuyLevels(iTs,'notfilled') == 0:
                        gain = self.cost2Prec(ts['symbol'],ts['costOut']-ts['costIn'])
                        self.message('Trading set %s on %s completed! Total gain: %s %s'%(ts['symbol'],self.exchange.name,gain,ts['baseCurrency']))
                        tradeSetsToDelete.append(iTs)
                self.unlockTradeSet(iTs)
        except Exception as e:
            # makes sure that the tradeSet deletion takes place even if some error occurred in another trade
            self.unlockTradeSet(iTs)
            raise(e)
        finally:  
            for iTs in tradeSetsToDelete:
                self.createTradeHistoryEntry(iTs)
                self.tradeSets.pop(iTs) 
    